"""Created on Aug 10 23:35:38 2024"""

from .generators import (
    multiple_models,
    multi_exponential,
    multi_base,
    multi_gaussian,
    multi_laplace,
    multi_gamma,
    multi_log_normal,
    multi_skew_normal,
    multi_half_normal,
    multi_folded_normal,
    multi_chi_squared,
)



"""Created on Aug 10 23:08:38 2024"""

import itertools
import warnings
from typing import Optional, Tuple, Union, List, Callable, Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from mpyez.backend.uPlotting import LinePlot  # type: ignore
from mpyez.ezPlotting import plot_xy  # type: ignore
from scipy.optimize import Bounds, curve_fit

# importing from files to avoid circular import
from .chiSquare_f import ChiSquareFitter
from .exponential_f import ExponentialFitter
from .foldedNormal_f import FoldedNormalFitter
from .gamma_f import GammaFitter
from .gaussian_f import GaussianFitter
from .halfNormal_f import HalfNormalFitter
from .laplace_f import LaplaceFitter
from .logNormal_f import LogNormalFitter
from .polynomial_f import LineFitter
from .skewNormal_f import SkewNormalFitter
from .utilities_f import sanity_check, _plot_fit
from .. import (
    epsilon,
    GAUSSIAN,
    LAPLACE,
    LINE,
    LOG_NORMAL,
    SKEW_NORMAL,
    CHI_SQUARE,
    EXPONENTIAL,
    FOLDED_NORMAL,
    GAMMA,
    NORMAL,
    HALF_NORMAL,
    ListOrNdArray,
    Params_)

# mock initialize the internal classes for auto MixedDataFitter class
fitter_dict = {
    CHI_SQUARE: ChiSquareFitter,
    EXPONENTIAL: ExponentialFitter,
    FOLDED_NORMAL: FoldedNormalFitter,
    GAMMA: GammaFitter,
    GAUSSIAN: GaussianFitter,
    NORMAL: GaussianFitter,
    HALF_NORMAL: HalfNormalFitter,
    LAPLACE: LaplaceFitter,
    LOG_NORMAL: LogNormalFitter,
    SKEW_NORMAL: SkewNormalFitter,
    LINE: LineFitter,
}


class MixedDataFitter:
    r"""
    Class to fit a mixture of different models to data.

    :param x_values: The x-values for the data.
    :param y_values: The y-values for the data.
    :param model_list: List of models to fit (e.g., `LINE`, `GAUSSIAN`, `LOG_NORMAL`)
    :param max_iterations: The maximum number of iterations for fitting procedure.
    """

    def __init__(
        self,
        x_values: ListOrNdArray,
        y_values: ListOrNdArray,
        model_list: List[str],
        fitter_dictionary: Optional[dict] = None,
        model_dictionary: Optional[dict] = None,
        max_iterations: int = 1000,
    ):
        # Check if the deprecated parameter was used
        if fitter_dictionary is not None:
            warnings.warn(
                message="`fitter_dictionary` is deprecated and will be removed in a future release. "
                        "Use `model_dictionary` instead.",
                category=DeprecationWarning,
                stacklevel=2
            )

        x_values, y_values = sanity_check(x_values=x_values, y_values=y_values)

        self.x_values: np.ndarray = x_values
        self.y_values: np.ndarray = y_values
        self.model_list = model_list
        self.max_iterations = max_iterations
        self.params: Any = None
        self.covariance: Any = None

        self.fitter_dict = fitter_dictionary or fitter_dict
        self.fitter_dict = model_dictionary or fitter_dict

        # self._validate_models()
        self.model_function = self._create_model_function()

    def __repr__(self):
        return (f"{self.__class__.__name__}(x_values={self.x_values}, y_values={self.y_values}, "
                f"model_list={self.model_list}, max_iterations={self.max_iterations})")

    def _create_model_function(self) -> Callable:
        """
        Creates a composite model function based on the specified models.

        :return: A composite model for fitting.
        """

        def _composite_model(x: np.ndarray, *params) -> np.ndarray:
            """
            Compute the composite model.

            Parameters
            ----------
            x : np.ndarray
                The x-values where the model is evaluated.
            params : tuple
                Parameters for the model components.

            Returns
            -------
            y : np.ndarray
                The computed y-values from the composite model.
            """
            y = np.zeros_like(a=x, dtype=float)
            param_index = 0

            for model in self.model_list:
                model_class = self._instantiate_class(model=model)
                n_par = self._instantiate_n_par(model=model)
                y += model_class.fitter(x=x, params=params[param_index: param_index + n_par])
                param_index += n_par

            return y

        return _composite_model

    def _expected_param_count(self) -> int:
        """
        Calculates the expected number of parameters based on the model list.

        :return: The number of parameters.
        """
        count = 0
        for model in self.model_list:
            count += self._instantiate_n_par(model=model)

        return count

    @staticmethod
    def _format_param(value, t_low=0.001, t_high=10_000) -> str:
        r"""
        Formats the parameter value to scientific notation based on its magnitude.

        Parameters
        ----------
        value: float
            The value of the parameter to be formatted.
        t_low: float, optional
            The lower bound below which the formatting should be applied to the value.
            Defaults to 0.001.
        t_high: float, optional
            The upper bound above which the formatting should be applied to the value.
            Defaults to 10,000.

        Returns
        -------
        str:
            A formatted string of the parameter value.
        """
        return f"{value:.3E}" if t_high < abs(value) or abs(value) < t_low else f"{value:.3f}"

    def _get_bounds(self):
        """
        Sets the bounds for each parameter based on the model list.

        :returns: Lower and upper bounds for the parameters.
        """
        lower_bounds = []
        upper_bounds = []

        for model in self.model_list:
            lb, ub = self._instantiate_bounds(model=model)
            lower_bounds.extend(lb)
            upper_bounds.extend(ub)

        return np.array(lower_bounds), np.array(upper_bounds)

    def _instantiate_class(self, model: str):
        try:
            fitter_instance = self.fitter_dict[model](x_values=np.array([]), y_values=np.array([]))
        except KeyError:
            raise ValueError(f"Model '{model}' not recognized. Ensure it is defined in the fitter dictionary.")

        return fitter_instance

    def _instantiate_n_par(self, model: str) -> int:
        return self._instantiate_class(model).n_par

    def _instantiate_bounds(self, model: str) -> Tuple[np.ndarray, np.ndarray]:
        return self._instantiate_class(model).fit_boundaries()

    def _parameter_extractor(self, values: np.ndarray) -> dict:
        """
        Extracts the parameters for each model in the model list.

        :param values: The values from which the model dictionary is to be extracted.

        :return: A dictionary where the keys are model names and the values are lists of parameters/error values.
        """
        p_index = 0
        param_dict: dict = {}

        for model in self.model_list:
            if model not in param_dict:
                param_dict[model] = []

            n_pars = self._instantiate_n_par(model=model)
            param_dict[model].extend([values[p_index: p_index + n_pars]])
            p_index += n_pars

        return param_dict

    def _params(self) -> np.ndarray:
        r"""
        Store the fitted parameters of the fitted model.

        Returns
        -------
        np.ndarray
            The parameters obtained after performing the fit.

        Raises
        ------
        RuntimeError
            If the fit has not been performed yet (i.e., ``self.params`` is ``None``).

        Notes
        -----
        This method assumes that the fitting process assigns values to ``self.params``.
        """
        if self.params is None:
            raise RuntimeError("Fit not performed yet. Call fit() first.")
        return self.params

    def _plot_individual_fitter(self, plotter):
        """
        Plot the individual fitters function.

        :param plotter: The plotting axis object
        """
        x = self.x_values
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"][1:]
        param_index = 0
        for i, model in enumerate(self.model_list):
            color = colors[i % len(colors)]
            class_model = self._instantiate_class(model=model)
            n_par = self._instantiate_n_par(model=model)
            pars = self.params[param_index: param_index + n_par]
            y_component = class_model.fitter(x=x, params=pars)
            plot_xy(
                x_data=x,
                y_data=y_component,
                x_label="",
                y_label="",
                plot_title="",
                data_label=f"{model.capitalize()} {i + 1}({', '.join(self._format_param(i) for i in pars)})",
                plot_dictionary=LinePlot(line_style="--", color=color),
                axis=plotter,
            )
            param_index += n_par

    def _standard_errors(self) -> np.ndarray:
        r"""
        Store the standard errors of the fitted parameters.

        Returns
        -------
        np.ndarray
            An array containing the standard errors of the fitted parameters.

        Raises
        ------
        RuntimeError
            If the fit has not been performed yet (i.e., ``self.covariance`` is ``None``).
        """
        if self.covariance is None:
            raise RuntimeError("Fit not performed yet. Call fit() first.")
        return np.sqrt(np.diag(self.covariance))

    def fit(self, p0: Params_, frozen: Optional[Union[int, List[int]]] = None):
        """
        Fit the data.

        :param p0: Initial guess for the fitted parameters.
        :type p0: Union[List[Tuple[int | float, ...]], np.ndarray]

        :param frozen: Parameter number of list of parameter numbers to freeze the value of.
        :type frozen: Union[int, List[int]]

        :raises ValueError: If the length of the initial guess is not equal to the expected parameter count.
        """
        p0_chain = p0.tolist() if isinstance(p0, np.ndarray) else p0

        # flatten cannot always work here because the mixed fitter might contain a variable number of parameters
        p0_chain = list(itertools.chain.from_iterable(p0_chain))
        if len(p0_chain) != self._expected_param_count():
            raise ValueError(f"Initial parameters length {len(p0_chain)} does not match expected count "
                             f"{self._expected_param_count()}.")

        lb, ub = self._get_bounds()

        if frozen:
            if isinstance(frozen, int):
                frozen = [frozen]
            for par_num in frozen:
                lb[par_num - 1] = p0_chain[par_num - 1] - epsilon
                ub[par_num - 1] = p0_chain[par_num - 1] + epsilon

        self.params, self.covariance, *_ = curve_fit(
            f=self.model_function,
            xdata=self.x_values,
            ydata=self.y_values,
            p0=np.array(p0_chain),
            maxfev=self.max_iterations,
            bounds=Bounds(lb=lb, ub=ub),
        )

    def get_fitted_curve(self) -> np.ndarray:
        """
        Gets the y-values from the fitted model.

        :return: The y-values from the fitted model

        :raises ValueError: If the model has not been fitted yet.
        """
        if self.params is None:
            raise RuntimeError("Fit not performed yet. Call fit() first.")

        return self.model_function(self.x_values, *self.params)

    def get_model_parameters(self, model: Optional[str] = None, errors: bool = False):
        """
        Extracts parameters (and error) values for a specific model, or for all models if no model is specified.

        :param model: Model name to extract parameters for. If unspecified, extracts parameters for all models.
            Defaults to ``None``.
        :param errors: If ``True``, includes the errors in the returned output. Defaults to ``False``.

        :return: A dictionary containing:

                - "parameters": Nested dictionary of parameter values for each model if `get_errors` is True.
                - "errors": Nested dictionary of errors for each model (if `get_errors=True`).

                Otherwise, returns just the parameters directly.
        """

        parameters = self._parameter_extractor(self.params)
        errs = self._parameter_extractor(np.sqrt(np.diag(self.covariance)))

        if not errors:
            return parameters if model is None else parameters.get(model, [])

        if model is None:
            # Return a combined dictionary for all models
            return {"parameters": parameters, "errors": errs}

        # Prepare output for a specific model
        output: dict = {"parameters": {}, "errors": {}}

        keys = ["parameters", "errors"]
        n_pars = self._instantiate_n_par(model=model)
        for temp_, key in zip([parameters, errs], keys):
            par_dict = temp_.get(model, [])
            if n_pars == 2:
                output[key] = par_dict
            else:
                output[key] = np.array_split(ary=np.asarray(a=par_dict, dtype=float).flatten(),
                                             indices_or_sections=n_pars)

        return output

    def get_value_error_pair(self, mean_values: bool = True, std_values: bool = False) -> np.ndarray:
        r"""
        Retrieve the value/error pairs for the fitted parameters.

        This method provides the fitted parameter values and their corresponding standard errors as a combined array or
        individually based on the input flags.

        Parameters
        ----------
        mean_values : bool, optional
            If ``True``, return only the values of the fitted parameters.
            Defaults to ``True``.
        std_values : bool, optional
            If ``True``, return only the standard errors of the fitted parameters.
            Defaults to ``False``.

        Returns
        -------
        np.ndarray
            - If ``mean_values`` and ``std_values`` are both ``True``: A 2D array of shape (n_parameters, 2),
                where each row is ``[value, error]``.
            - If ``mean_values`` is ``True`` and ``std_values`` is ``False``: A 1D array of parameter values.
            - If ``std_values`` is ``True`` and ``mean_values`` is ``False``: A 1D array of standard errors.
            - If both flags are ``False``: An error message.

        Raises
        ------
        ValueError
            If both ``mean_values`` and ``std_values`` are ``False``.
        """
        pairs = np.column_stack([self._params(), self._standard_errors()])

        if mean_values and std_values:
            return pairs
        elif mean_values:
            return pairs[:, 0]
        elif std_values:
            return pairs[:, 1]
        else:
            raise ValueError("Either 'mean_values' or 'std_values' must be True.")

    def plot_fit(
        self,
        show_individuals: bool = False,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        data_label: Optional[str] = None,
        fit_label: Optional[str] = None,
        title: Optional[str] = None,
        axis: Optional[Axes] = None,
    ):
        """
        Plot the fitted models.

        Parameters
        ----------
        show_individuals: bool, optional
            Whether to show individually fitted models or not.
        x_label: str, optional
            The label for the x-axis.
        y_label: str, optional
            The label for the y-axis.
        title: str, optional
            The title for the plot.
        data_label: str, optional
            The label for the data.
        fit_label: str, optional
            The label for the fitted model.
        axis: Axes, optional
            Axes to plot instead of the entire figure. Defaults to None.

        Returns
        -------
        plotter
            The plotter handle for the drawn plot.
        """
        return _plot_fit(
            x_values=self.x_values,
            y_values=self.y_values,
            parameters=self.params,
            n_fits=len(self.model_list),
            class_name=self.__class__.__name__,
            _n_fitter=self.model_function,
            _n_plotter=self._plot_individual_fitter,
            show_individuals=show_individuals,
            x_label=x_label,
            y_label=y_label,
            title=title,
            data_label=data_label,
            fit_label=fit_label,
            axis=axis,
        )



"""Created on Feb 02 03:46:43 2025"""

from typing import Dict

import numpy as np

from ..backend import BaseDistribution, errorHandling as erH
from ..utilities_d import (
    scaled_inv_chi_square_pdf_,
    scaled_inv_chi_square_log_pdf_,
    scaled_inv_chi_square_cdf_,
    scaled_inv_chi_square_log_cdf_,
)
from ... import md_scipy_like


class ScaledInverseChiSquareDistribution(BaseDistribution):
    def __init__(
        self,
        amplitude: float = 1.0,
        df: float = 1.0,
        scale: float = 1.0,
        loc: float = 0.0,
        normalize: bool = False,
    ):
        if not normalize and amplitude < 0:
            raise erH.NegativeAmplitudeError()
        if scale < 0:
            raise erH.NegativeScaleError()

        self.amplitude = 1 if normalize else amplitude
        self.df = df
        self.scale = scale
        self.tau2 = scale / df

        self.loc = loc
        self.norm = normalize

    @classmethod
    @md_scipy_like('1.0.7')
    def scipy_like(cls, a, loc: float = 0.0, scale=1.0):
        return cls(df=a, loc=loc, scale=scale, normalize=True)

    @classmethod
    def from_scipy_params(cls, a, loc: float = 0.0, scale=1.0):
        return cls(df=a, loc=loc, scale=scale, normalize=True)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return scaled_inv_chi_square_pdf_(
            x,
            amplitude=self.amplitude,
            df=self.df,
            scale=self.scale,
            loc=self.loc,
            normalize=self.norm,
        )

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        return scaled_inv_chi_square_log_pdf_(
            x,
            amplitude=self.amplitude,
            df=self.df,
            scale=self.scale,
            loc=self.loc,
            normalize=self.norm,
        )

    def cdf(self, x: np.ndarray) -> np.ndarray:
        return scaled_inv_chi_square_cdf_(
            x,
            amplitude=self.amplitude,
            df=self.df,
            scale=self.scale,
            loc=self.loc,
            normalize=self.norm,
        )

    def logcdf(self, x: np.ndarray) -> np.ndarray:
        return scaled_inv_chi_square_log_cdf_(
            x,
            amplitude=self.amplitude,
            df=self.df,
            loc=self.loc,
            scale=self.scale,
            normalize=self.norm,
        )

    def stats(self) -> Dict[str, float]:
        v, tau2, loc = self.df, self.tau2, self.loc
        mean_ = (v * tau2) / (v - 2)
        mode_ = (v * tau2) / (v + 2)
        variance_ = (2 * v**2 * tau2**2) / ((v - 2)**2 * (v - 4))

        return {
            "mean": mean_ + loc if v > 2 else np.inf,
            "mode": mode_ + loc,
            "variance": variance_ if v > 4 else np.inf,
            "std": np.sqrt(variance_) if v > 4 else np.inf,
        }



"""Created on Jul 18 00:15:42 2024"""

from typing import Union, Sequence, Tuple, List

import numpy as np
from deprecated.sphinx import deprecated

from .version import (
    __author__,
    __copyright__,
    __description__,
    __email__,
    __license__,
    __url__,
    __version__,
)


def mark_deprecated(ver_: str, new: str):
    """Decorator that marks a scipy_like-style method as deprecated.

    Automatically extracts the method name and constructs a standardized warning.

    Parameters
    ----------
    ver_ : str
        The version where the method is deprecated.
    new : str
        The name of the method to use instead.
    """

    def decorator(func):
        method_name = func.__name__
        reason = f"Use `{new}` instead of `{method_name}`. `{method_name}` will be removed in a future release."
        return deprecated(version=ver_, reason=reason)(func)

    return decorator


def md_scipy_like(ver_: str, new: str = 'from_scipy_params'):
    return mark_deprecated(ver_, new)


doc_style = "numpy_napoleon_with_merge"

INF = np.inf
LOG = np.log

# taken from https://stackoverflow.com/a/19141711
EPSILON = np.finfo(float).eps
epsilon = np.sqrt(EPSILON)

ListOrNdArray = Union[List[int | float], np.ndarray]

ParamTuple = Tuple[int | float, ...]
Params_ = Union[List[ParamTuple], np.ndarray]

GAUSSIAN = "gaussian"
NORMAL = GAUSSIAN

ARC_SINE = "arc_sine"
BETA = "beta"
CHI_SQUARE = "chi_square"
EXPONENTIAL = "exponential"
FOLDED_NORMAL = "folded_normal"
GAMMA = "gamma"
HALF_NORMAL = "half_normal"
LAPLACE = "laplace"
LOG_NORMAL = "log_normal"
SKEW_NORMAL = "skew_normal"

LINE = "line"
LINEAR = LINE
QUADRATIC = "quad"
CUBIC = "cubic"



"""Created on Jan 29 15:42:23 2025"""

from typing import Dict

import numpy as np
from scipy.special import gammaln

from ..backend import BaseDistribution, errorHandling as erH
from ..utilities_d import sym_gen_normal_pdf_, sym_gen_normal_cdf_
from ... import md_scipy_like


class SymmetricGeneralizedNormalDistribution(BaseDistribution):
    r"""
    Class for SymmetricGeneralizedNormalDistribution.

    :param amplitude: The amplitude of the PDF. Defaults to 1.0. Ignored if **normalize** is ``True``.
    :type amplitude: float, optional

    :param shape: The shape parameter, :math:`\beta`. Defaults to 1.0.
    :type shape: float, optional

    :param loc: The shape parameter, :math:`\mu`. Defaults to 0.0.
    :type loc: float, optional

    :param scale: The standard deviation parameter, :math:`\alpha`. Defaults to 1.0.
    :type scale: float, optional

    :param normalize: If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
     Defaults to ``False``.
    :type normalize: bool, optional

    :raise NegativeAmplitudeError: If the provided value of amplitude is negative.
    :raise NegativeScaleError: If the provided value of scale parameter is negative.

    Examples
    --------
    Importing libraries:

    .. literalinclude:: ../../../examples/basic/gaussian.py
       :language: python
       :linenos:
       :lineno-start: 3
       :lines: 3-7

    Generating a standard SymmetricGeneralizedNormalDistribution(:math:`\beta=1, \mu=0, \alpha = 1`)
     with ``pyMultiFit`` and ``scipy``:

    .. literalinclude:: ../../../examples/basic/gennorm.py
       :language: python
       :linenos:
       :lineno-start: 9
       :lines: 9-12

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/gennorm.py
       :language: python
       :linenos:
       :lineno-start: 14
       :lines: 14-29

    .. image:: ../../../images/gen_norm_example1.png
       :alt: GenNorm(1, 0, 1)
       :align: center

    Generating a scaled and translated SymmetricGeneralizedNormalDistribution(:math:`\beta=2, \mu=-3, \alpha=5`):

    .. literalinclude:: ../../../examples/basic/gennorm.py
       :language: python
       :lineno-start: 32
       :lines: 32

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/gennorm.py
       :language: python
       :lineno-start: 34
       :lines: 34-49

    .. image:: ../../../images/gen_norm_example2.png
       :alt: GenNorm(2, -3, 5)
       :align: center
    """

    def __init__(
        self,
        amplitude: float = 1.0,
        shape: float = 1.0,
        loc: float = 0.0,
        scale: float = 1.0,
        normalize: bool = False,
    ):
        if amplitude < 0 and not normalize:
            raise erH.NegativeAmplitudeError()
        if shape < 0:
            raise erH.NegativeShapeError()
        self.amplitude = 1.0 if normalize else amplitude
        self.loc = loc
        self.scale = scale
        self.shape = shape

        self.norm = normalize

    @classmethod
    @md_scipy_like('1.0.7')
    def scipy_like(cls, beta, loc: float = 0.0, scale: float = 1.0):
        """
        Instantiate SymmetricGeneralizedNormalDistribution with scipy parametrization.

        Parameters
        ----------
        beta: float
            The shape parameter.
        loc: float, optional
            The mean parameter. Defaults to 0.0.
        scale: float, optional
            The scale parameter. Defaults to 1.0.

        Returns
        -------
        SymmetricGeneralizedNormalDistribution
            An instance of normalized SymmetricGeneralizedNormalDistribution.
        """
        return cls(shape=beta, loc=loc, scale=scale, normalize=True)

    @classmethod
    def from_scipy_params(cls, beta, loc: float = 0.0, scale: float = 1.0):
        """
        Instantiate SymmetricGeneralizedNormalDistribution with scipy parametrization.

        Parameters
        ----------
        beta: float
            The shape parameter.
        loc: float, optional
            The mean parameter. Defaults to 0.0.
        scale: float, optional
            The scale parameter. Defaults to 1.0.

        Returns
        -------
        SymmetricGeneralizedNormalDistribution
            An instance of normalized SymmetricGeneralizedNormalDistribution.
        """
        return cls(shape=beta, loc=loc, scale=scale, normalize=True)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return sym_gen_normal_pdf_(
            x,
            amplitude=self.amplitude,
            shape=self.shape,
            loc=self.loc,
            scale=self.scale,
            normalize=self.norm,
        )

    def cdf(self, x: np.ndarray) -> np.ndarray:
        return sym_gen_normal_cdf_(
            x,
            amplitude=self.amplitude,
            shape=self.shape,
            loc=self.loc,
            scale=self.scale,
            normalize=self.norm,
        )

    def stats(self) -> Dict[str, float]:
        mean_ = self.loc
        median_ = self.loc
        mode_ = self.loc

        variance_ = 2 * np.log(self.scale) + gammaln(3 / self.shape) - gammaln(1 / self.shape)
        variance_ = np.exp(variance_)

        return {
            "mean": mean_,
            "median": median_,
            "mode": mode_,
            "variance": variance_,
            "std": np.sqrt(variance_),
        }



"""Created on Aug 03 20:07:50 2024"""

from typing import Dict

import numpy as np

from .backend import BaseDistribution, errorHandling as erH
from .utilities_d import gaussian_cdf_, gaussian_pdf_, gaussian_log_pdf_, gaussian_log_cdf_
from .. import md_scipy_like


class GaussianDistribution(BaseDistribution):
    r"""
    Class for Gaussian distribution.

    :param amplitude: The amplitude of the PDF. Defaults to 1.0. Ignored if **normalize** is ``True``.
    :type amplitude: float, optional

    :param mu: The mean parameter, :math:`\mu`. Defaults to 0.0.
    :type mu: float, optional

    :param std: The standard deviation parameter, :math:`\sigma`. Defaults to 1.0.
    :type std: float, optional

    :param normalize: If ``True``, the distribution is normalized so that the total area under the PDF equals 1. Defaults to ``False``.
    :type normalize: bool, optional

    :raise NegativeAmplitudeError: If the provided value of amplitude is negative.
    :raise NegativeStandardDeviationError: If the provided value of standard deviation is negative.

    Examples
    --------
    Importing libraries:

    .. literalinclude:: ../../../examples/basic/gaussian.py
       :language: python
       :linenos:
       :lineno-start: 3
       :lines: 3-7

    Generating a standard Gaussian(:math:`\mu=0, \sigma = 1`) distribution with ``pyMultiFit`` and ``scipy``:

    .. literalinclude:: ../../../examples/basic/gaussian.py
       :language: python
       :linenos:
       :lineno-start: 9
       :lines: 9-12

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/gaussian.py
       :language: python
       :linenos:
       :lineno-start: 14
       :lines: 14-29

    .. image:: ../../../images/gaussian_example1.png
       :alt: Gaussian(0, 1)
       :align: center

    Generating a translated Gaussian(:math:`\mu=3, \sigma=2`) distribution:

    .. literalinclude:: ../../../examples/basic/gaussian.py
       :language: python
       :lineno-start: 32
       :lines: 32

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/gaussian.py
       :language: python
       :lineno-start: 34
       :lines: 34-49

    .. image:: ../../../images/gaussian_example2.png
       :alt: Gaussian(3, 2)
       :align: center
    """

    def __init__(
        self,
        amplitude: float = 1.0,
        mu: float = 0.0,
        std: float = 1.0,
        normalize: bool = False,
    ):
        if not normalize and amplitude <= 0:
            raise erH.NegativeAmplitudeError()
        if std <= 0:
            raise erH.NegativeStandardDeviationError()

        self.amplitude = 1.0 if normalize else amplitude
        self.mu = mu
        self.std_ = std
        self.norm = normalize

    @classmethod
    @md_scipy_like('1.0.7')
    def scipy_like(cls, loc: float = 0.0, scale: float = 1.0) -> 'GaussianDistribution':
        """
        Instantiate GaussianDistribution with scipy parametrization.

        Parameters
        ----------
        loc: float, optional
            The mean parameter. Defaults to 0.0.
        scale: float, optional
            The scale parameter. Defaults to 1.0.

        Returns
        -------
        GaussianDistribution
            An instance of normalized GaussianDistribution.
        """
        return cls(mu=loc, std=scale, normalize=True)

    @classmethod
    def from_scipy_params(cls, loc: float = 0.0, scale: float = 1.0) -> 'GaussianDistribution':
        """
        Instantiate GaussianDistribution with scipy parametrization.

        Parameters
        ----------
        loc: float, optional
            The mean parameter. Defaults to 0.0.
        scale: float, optional
            The scale parameter. Defaults to 1.0.

        Returns
        -------
        GaussianDistribution
            An instance of normalized GaussianDistribution.
        """
        return cls(mu=loc, std=scale, normalize=True)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return gaussian_pdf_(
            x,
            amplitude=self.amplitude,
            mean=self.mu,
            std=self.std_,
            normalize=self.norm,
        )

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        return gaussian_log_pdf_(
            x,
            amplitude=self.amplitude,
            mean=self.mu,
            std=self.std_,
            normalize=self.norm,
        )

    def cdf(self, x: np.ndarray) -> np.ndarray:
        return gaussian_cdf_(
            x,
            amplitude=self.amplitude,
            mean=self.mu,
            std=self.std_,
            normalize=self.norm,
        )

    def logcdf(self, x: np.ndarray) -> np.ndarray:
        return gaussian_log_cdf_(
            x,
            amplitude=self.amplitude,
            mean=self.mu,
            std=self.std_,
            normalize=self.norm,
        )

    def stats(self) -> Dict[str, float]:
        m, s = self.mu, self.std_

        return {
            "mean": m,
            "median": m,
            "mode": m,
            "variance": s**2,
            "std": s,
        }



"""Created on Dec 04 23:24:55 2024"""

import numpy as np

from .backend import BaseFitter
from .utilities_f import sanity_check
from .. import ListOrNdArray, Params_
from ..distributions.utilities_d import folded_normal_pdf_


class FoldedNormalFitter(BaseFitter):
    """A class for fitting multiple FoldedNormal distributions to the given data."""

    def __init__(
        self,
        x_values: ListOrNdArray,
        y_values: ListOrNdArray,
        max_iterations: int = 1000,
    ):
        x_values, y_values = sanity_check(x_values=x_values, y_values=y_values)
        super().__init__(x_values=x_values, y_values=y_values, max_iterations=max_iterations)
        self.n_par = 3

    def fit_boundaries(self):
        lb = [0, -np.inf, 0]
        ub = [np.inf, np.inf, np.inf]
        return lb, ub

    @staticmethod
    def fitter(x, params: Params_):
        return folded_normal_pdf_(x, *params)



"""Created on Aug 14 01:28:13 2024"""

from typing import Dict

import numpy as np

from .backend import BaseDistribution, errorHandling as erH
from .utilities_d import (
    gamma_pdf_,
    gamma_log_pdf_,
    gamma_cdf_,
    gamma_log_cdf_,
)
from .. import md_scipy_like


class GammaDistribution(BaseDistribution):
    r"""
    Class for Gamma distribution with shape and scale parameters.

    :param amplitude: The amplitude of the PDF. Default is 1.0. Ignored if **normalize** is ``True``.
    :type amplitude: float, optional

    :param shape: The shape parameter, :math:`\alpha`. Defaults to 1.0.
    :type shape: float, optional

    :param scale: The rate parameter, :math:`\theta`. Defaults to 1.0.
    :type scale: float, optional

    :param loc: The location parameter, for shifting. Defaults to 0.0.
    :type loc: float, optional

    :param normalize: If ``True``, the distribution is normalized so that the total area under the PDF equals 1. Defaults to ``False``.
    :type normalize: bool, optional

    :raise NegativeAmplitudeError: If the provided value of amplitude is negative.
    :raise NegativeShapeError: If the provided value of shape is negative.
    :raise NegativeScaleError: If the provided value of scale is negative.

    Examples
    --------
    Importing libraries:

    .. literalinclude:: ../../../examples/basic/gamma_.py
       :language: python
       :linenos:
       :lineno-start: 3
       :lines: 3-7

    Generating a standard GammaSS(:math:`\alpha =1.5, \lambda = 1`) distribution with ``pyMultiFit`` and ``scipy``:

    .. literalinclude:: ../../../examples/basic/gamma_.py
       :language: python
       :linenos:
       :lineno-start: 9
       :lines: 9-12

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/gamma_.py
       :language: python
       :linenos:
       :lineno-start: 14
       :lines: 14-29

    .. image:: ../../../images/gammaSS_example1.png
       :alt: GammaSS(1.5, 1)
       :align: center

    Generating a translated Gamma(:math:`\alpha=1.5, \lambda=0.2`) distribution with :math:`\text{loc} = 3`:

    .. literalinclude:: ../../../examples/basic/gamma_.py
       :language: python
       :lineno-start: 32
       :lines: 32

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/gamma_.py
       :language: python
       :lineno-start: 34
       :lines: 34-49

    .. image:: ../../../images/gammaSS_example2.png
       :alt: GammaSS(1.5, 0.2, 3)
       :align: center
    """

    def __init__(
        self,
        amplitude: float = 1.0,
        shape: float = 1.0,
        scale: float = 1.0,
        loc: float = 0.0,
        normalize: bool = False,
    ):
        if not normalize and amplitude <= 0:
            raise erH.NegativeAmplitudeError()
        if shape <= 0:
            raise erH.NegativeShapeError()
        if scale <= 0:
            raise erH.NegativeScaleError()
        self.amplitude = 1.0 if normalize else amplitude
        self.shape = shape
        self.scale = scale
        self.loc = loc

        self.norm = normalize

    @classmethod
    @md_scipy_like('1.0.7')
    def scipy_like(cls, a: float, loc: float = 0.0, scale: float = 1.0) -> 'GammaDistribution':
        r"""
        Instantiate GammaDistributionSS with scipy parametrization.

        Parameters
        ----------
        a: float
            The shape parameter.
        loc: float, optional
            The location parameter. Defaults to 0.0.
        scale: float, optional
            The scaling parameter. Defaults to 1.0.

        Returns
        -------
        GammaDistribution
            An instance of normalized GammaDistributionSS.
        """
        return cls(shape=a, loc=loc, scale=scale, normalize=True)

    @classmethod
    def from_scipy_params(cls, a: float, loc: float = 0.0, scale: float = 1.0) -> 'GammaDistribution':
        r"""
        Instantiate GammaDistributionSS with scipy parametrization.

        Parameters
        ----------
        a: float
            The shape parameter.
        loc: float, optional
            The location parameter. Defaults to 0.0.
        scale: float, optional
            The scaling parameter. Defaults to 1.0.

        Returns
        -------
        GammaDistribution
            An instance of normalized GammaDistributionSS.
        """
        return cls(shape=a, loc=loc, scale=scale, normalize=True)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return gamma_pdf_(x, amplitude=self.amplitude, alpha=self.shape, theta=self.scale, loc=self.loc,
                          normalize=self.norm)

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        return gamma_log_pdf_(x, amplitude=self.amplitude, alpha=self.shape, theta=self.scale, loc=self.loc,
                              normalize=self.norm)

    def cdf(self, x: np.ndarray) -> np.ndarray:
        return gamma_cdf_(x, amplitude=self.amplitude, alpha=self.shape, theta=self.scale, loc=self.loc,
                          normalize=self.norm)

    def logcdf(self, x: np.ndarray) -> np.ndarray:
        return gamma_log_cdf_(x, amplitude=self.amplitude, alpha=self.shape, theta=self.scale, loc=self.loc,
                              normalize=self.norm)

    def stats(self) -> Dict[str, float]:
        s, r, l_ = self.shape, self.scale, self.loc

        mean_ = (s * r) + l_
        variance_ = s * r**2
        mode_ = (s - 1) * r + l_ if s >= 1 else 0

        return {
            "mean": mean_,
            "mode": mode_,
            "variance": variance_,
            "std": np.sqrt(variance_),
        }



"""Created on Dec 04 03:21:09 2024"""

import numpy as np

from .backend import BaseFitter
from .utilities_f import sanity_check
from .. import ListOrNdArray, Params_
from ..distributions.utilities_d import chi_square_pdf_


class ChiSquareFitter(BaseFitter):
    """A class for fitting multiple ChiSquare distributions to the given data."""

    def __init__(
        self,
        x_values: ListOrNdArray,
        y_values: ListOrNdArray,
        max_iterations: int = 1000,
    ):
        x_values, y_values = sanity_check(x_values=x_values, y_values=y_values)
        super().__init__(x_values=x_values, y_values=y_values, max_iterations=max_iterations)
        self.n_par = 3
        self.pn_par = 2
        self.sn_par = {"loc": 0.0}

    def fit_boundaries(self):
        lb = (0, 0, -np.inf)
        ub = (np.inf, np.inf, np.inf)
        return lb, ub

    @staticmethod
    def fitter(x, params: Params_):
        return chi_square_pdf_(x, *params)



"""Created on Aug 03 17:13:21 2024"""

__all__ = [
    "_beta_expr",
    "_pdf_scaling",
    "preprocess_input",
    "arc_sine_pdf_",
    "arc_sine_cdf_",
    "arc_sine_log_pdf_",
    "arc_sine_log_cdf_",
    "beta_pdf_",
    "beta_cdf_",
    "beta_log_pdf_",
    "beta_log_cdf_",
    "chi_square_pdf_",
    "chi_square_cdf_",
    "chi_square_log_pdf_",
    "chi_square_log_cdf_",
    "exponential_pdf_",
    "exponential_cdf_",
    "exponential_log_pdf_",
    "exponential_log_cdf_",
    "folded_normal_pdf_",
    "folded_normal_cdf_",
    "folded_normal_log_pdf_",
    "folded_normal_log_cdf_",
    "gamma_pdf_",
    "gamma_log_pdf_",
    "gamma_cdf_",
    "gamma_log_cdf_",
    "sym_gen_normal_pdf_",
    "sym_gen_normal_cdf_",
    "sym_gen_normal_log_pdf_",
    "sym_gen_normal_log_cdf_",
    "gaussian_pdf_",
    "gaussian_cdf_",
    "gaussian_log_pdf_",
    "gaussian_log_cdf_",
    "half_normal_pdf_",
    "half_normal_cdf_",
    "half_normal_log_pdf_",
    "half_normal_log_cdf_",
    "laplace_pdf_",
    "laplace_cdf_",
    "laplace_log_pdf_",
    "laplace_log_cdf_",
    "log_normal_pdf_",
    "log_normal_cdf_",
    "log_normal_log_pdf_",
    "log_normal_log_cdf_",
    "scaled_inv_chi_square_pdf_",
    "scaled_inv_chi_square_log_pdf_",
    "scaled_inv_chi_square_cdf_",
    "scaled_inv_chi_square_log_cdf_",
    "skew_normal_pdf_",
    "skew_normal_cdf_",
    "uniform_pdf_",
    "uniform_cdf_",
    "uniform_log_pdf_",
    "uniform_log_cdf_",
    "line",
    "quadratic",
    "cubic",
]

import functools
from typing import Union, Callable

import numpy as np
import scipy.special as ssp
from custom_inherit import doc_inherit

from .. import doc_style, LOG, INF, ListOrNdArray

TWO = 2.0
SQRT_TWO = np.sqrt(TWO)
LOG_TWO = LOG(TWO)
LOG_SQRT_TWO = ssp.xlogy(0.5, TWO)

PI = np.pi
SQRT_PI = np.sqrt(PI)
LOG_PI = LOG(PI)
LOG_SQRT_PI = ssp.xlogy(0.5, PI)

TWO_PI = 2 * PI
SQRT_TWO_PI = np.sqrt(TWO_PI)
LOG_TWO_PI = LOG(TWO_PI)
LOG_SQRT_TWO_PI = ssp.xlogy(0.5, TWO_PI)

INV_PI = 1.0 / PI
TWO_BY_PI = 2.0 * INV_PI
SQRT_TWO_BY_PI = np.sqrt(TWO_BY_PI)
LOG_TWO_BY_PI = LOG(TWO_BY_PI)
LOG_SQRT_TWO_BY_PI = ssp.xlogy(0.5, TWO_BY_PI)


def suppress_numpy_warnings():
    """
    A decorator that suppresses NumPy warnings using np.errstate.

    Parameters (all optional):
        divide, over, under, invalid: Can be 'ignore', 'warn', 'raise', 'call', 'print', or 'log'
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with np.errstate(all="ignore"):
                return func(*args, **kwargs)

        return wrapper

    return decorator


@suppress_numpy_warnings()
def arc_sine_pdf_(
    x: ListOrNdArray, amplitude: float = 1.0, loc: float = 0.0, scale: float = 1.0, normalize: bool = False
) -> np.ndarray:
    r"""
    Compute PDF of :class:`~pymultifit.distributions.arcSine_d.ArcSineDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values where PDF is evaluated.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    loc : float, optional
        The location parameter specifying the lower bound of the distribution.
        Defaults to 0.0.
    scale : float, optional
        The scale parameter, specifying the width of the distribution.
        Defaults to 1.0.
    normalize : bool, optional
        If True, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The ArcSine PDF is defined as:

    .. math:: f(y) = \frac{1}{\pi \sqrt{y(1-y)}}

    where, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final PDF is expressed as :math:`f(y)/\text{scale}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    if scale < 0:
        return np.full(y.shape, np.nan)

    c1 = (y > 0) & (y < 1)
    c2 = y == 0
    c3 = y == 1

    z = y * (1 - y)

    pdf_ = np.select(condlist=[c1, c2, c3], choicelist=[1 / PI / np.sqrt(z), INF, INF], default=0.0)
    pdf_ /= scale

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=arc_sine_pdf_, style=doc_style)
def arc_sine_log_pdf_(
    x: ListOrNdArray, amplitude: float = 1.0, loc: float = 0.0, scale: float = 1.0, normalize: bool = False
) -> np.ndarray:
    r"""
    Compute logPDF of :class:`~pymultifit.distributions.arcSine_d.ArcSineDistribution`.

    Notes
    -----
    The ArcSine logPDF is defined as:

    .. math:: \ell(y) = -\ln(\pi)  - 0.5\ln(y-y^2)

    where, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final logPDF is expressed as :math:`\ell(y) - \ln(\text{scale})`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    c1 = (y > 0) & (y < 1)
    c2 = y == 0
    c3 = y == 1

    z = y * (1 - y)

    log_pdf_ = np.select(condlist=[c1, c2, c3], choicelist=[LOG(1 / PI / np.sqrt(z)), INF, INF], default=-INF)
    log_pdf_ -= LOG(scale)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=arc_sine_pdf_, style=doc_style)
def arc_sine_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    if scale < 0:
        return np.full(y.shape, np.nan)

    c1 = (y > 0) & (y < 1)
    c2 = y < 1

    return np.select(condlist=[c1, c2], choicelist=[TWO_BY_PI * np.arcsin(np.sqrt(y)), 0.0], default=1.0)


@suppress_numpy_warnings()
@doc_inherit(parent=arc_sine_cdf_, style=doc_style)
def arc_sine_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF of :class:`~pymultifit.distributions.arcSine_d.ArcSineDistribution`.

    Notes
    -----
    The ArcSine log CDF is defined as:

    .. math:: \mathcal{L}(y) = \ln\left(\frac{2}{\pi}\right) + \ln\arcsin(\sqrt{y})

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}

    The final logCDF is expressed as :math:`\mathcal{L}(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    c1 = (y > 0) & (y < 1)
    c2 = y < 1

    return np.select(condlist=[c1, c2], choicelist=[LOG(TWO_BY_PI * np.arcsin(np.sqrt(y))), -INF], default=0.0)


@suppress_numpy_warnings()
def beta_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    beta_: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF of :class:`~pymultifit.distributions.beta_d.BetaDistribution`.

    Parameters
    ----------
    beta_
    x : np.ndarray
        Input array of values where PDF is evaluated.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    alpha : float, optional
        The :math:`\alpha` parameter.
        Default is 1.0.
    beta_ : float, optional
        The :math:`\beta` parameter.
        Default is 1.0.
    loc : float, optional
        The location parameter, for shifting.
        Default is 0.0.
    scale : float, optional
        The scale parameter, for scaling.
        Default is 1.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as `x`, containing the evaluated values.

    Notes
    -----
    The Beta PDF is defined as:

    .. math:: f(y; \alpha, \beta) = \frac{y^{\alpha - 1} (1 - y)^{\beta - 1}}{B(\alpha, \beta)}

    where :math:`B(\alpha, \beta)` is the Beta function (see, :obj:`ssp.beta`), and :math:`y` is the
    transformed value of :math:`x` such that:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final PDF is expressed as :math:`f(y)/\text{scale}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    conditions, main = _beta_expr(y=y, a=alpha, b=beta_, un_log=True)

    pdf_ = np.select(condlist=conditions, choicelist=[1, np.nan, main], default=0)
    pdf_ /= scale

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=beta_pdf_, style=doc_style)
def beta_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    beta_: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""Compute logPDF for :class:`~pymultifit.distributions.beta_d.BetaDistribution`.

    Notes
    -----
    The Beta logPDFis defined as

    .. math:: \ell(y) = (\alpha - 1)\ln(y) + (\beta - 1)\ln(1 - y) - \ln(\text{Beta}(\alpha, \beta))

    where :math:`B(\alpha, \beta)` is the :obj:`~ssp.beta` function, and :math:`y` is the
    transformed value of :math:`x` such that:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final logPDF is expressed as :math:`\ell(y) - \ln(\text{scale})`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    conditions, main = _beta_expr(y=y, a=alpha, b=beta_)

    log_pdf_ = np.select(condlist=conditions, choicelist=[0.0, np.nan, main], default=-INF)
    log_pdf_ -= LOG(scale)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=beta_pdf_, style=doc_style)
def beta_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    beta_: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF for :class:`~pymultifit.distributions.beta_d.BetaDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        For API consistency only.
    normalize : bool, optional
        For API consistency only.

    Notes
    -----
    The Beta CDF is defined as:

    .. math:: F(y) = I_y(\alpha, \beta)

    where :math:`I_y(\alpha, \beta)` is the :obj:`~ssp.betainc` function, and :math:`y` is the transformed
    value of :math:`x`, defined as:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    return np.select(condlist=[y > 1, y < 0], choicelist=[1, 0], default=ssp.betainc(alpha, beta_, y))


@suppress_numpy_warnings()
@doc_inherit(parent=beta_cdf_, style=doc_style)
def beta_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    beta_: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute logCDF for :class:`~pymultifit.distributions.beta_d.BetaDistribution`.

    Notes
    -----
    The Beta logCDF is defined as:

    .. math:: \mathcal{L}(y) = \ln I_y(\alpha, \beta)

    where :math:`I_y(\alpha, \beta)` is the :obj:`~ssp.betainc` function, and :math:`y` is the transformed
    value of :math:`x`, defined as:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final logCDF is expressed as :math:`\mathcal{L}(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    return np.select(condlist=[y > 1, y < 0], choicelist=[0, -INF], default=LOG(ssp.betainc(alpha, beta_, y)))


@suppress_numpy_warnings()
def chi_square_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    degree_of_freedom: Union[int, float] = 1,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF for :mod:`~pymultifit.distributions.chiSquare_d.ChiSquareDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    degree_of_freedom : int, optional
        The degrees of freedom parameter.
        Defaults to 1.
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    scale: float, optional
        The scale parameter, for scaling.
        Defaults to 1.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The ChiSquare PDF is defined as:

    .. math:: f(y\ |\ k) = \dfrac{y^{(k/2) - 1} e^{-y/2}}{2^{k/2} \Gamma(k/2)}

    where :math:`\Gamma(\cdot)` is the :obj:`~ssp.gamma` function, and :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}

    The final PDF is expressed as :math:`f(y)/\text{scale}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    df_half = degree_of_freedom / 2.0

    pdf_ = np.where(y > 0, np.exp(_chi2(y=y, df_half=df_half)), 0)
    pdf_ /= scale

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


def _chi2(y, df_half):
    return ssp.xlogy(df_half - 1, y) - (y / 2) - ssp.gammaln(df_half) - (LOG_TWO * df_half)


@suppress_numpy_warnings()
@doc_inherit(parent=chi_square_pdf_, style=doc_style)
def chi_square_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    degree_of_freedom: Union[int, float] = 1,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF for :mod:`~pymultifit.distributions.chiSquare_d.ChiSquareDistribution`.

    Notes
    -----
    The ChiSquare log PDF is defined as:

    .. math:: \ell(y\ |\ k) = \left(\dfrac{k}{2} - 1\right)\ln(y) - \dfrac{y}{2} - \dfrac{k}{2}\ln(2) - \ln\Gamma\left(\dfrac{k}{2}\right)

    where :math:`\ln\Gamma(\cdot)` is the :obj:`~ssp.gammaln` function,
    and :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}

    The final PDF is expressed as :math:`\ell(y) - \ln(\text{scale})`.

    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    df_half = degree_of_freedom / 2.0

    log_pdf_ = np.where(y > 0, _chi2(y=y, df_half=df_half), -INF)
    log_pdf_ -= LOG(scale)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=chi_square_pdf_, style=doc_style)
def chi_square_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    degree_of_freedom: Union[int, float] = 1,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF for :mod:`~pymultifit.distributions.chiSquare_d.ChiSquareDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: bool, optional
        For API consistency only.
    Notes
    -----
    The ChiSquare CDF is defined as:

    .. math:: F(y) = \gamma\left(\dfrac{\nu}{2}, \dfrac{y}{2}\right)

    where, :math:`\gamma\left(\cdot, \cdot\right)` is the :obj:`~ssp.gammainc` lower regularized incomplete
    gamma function, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}

    The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    return np.where(y > 0, ssp.gammainc(degree_of_freedom / 2, y / 2), 0)


@suppress_numpy_warnings()
@doc_inherit(parent=chi_square_cdf_, style=doc_style)
def chi_square_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    degree_of_freedom: Union[int, float] = 1,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF for :mod:`~pymultifit.distributions.chiSquare_d.ChiSquareDistribution`.

    Notes
    -----
    The ChiSquare logCDF is defined as:

    .. math:: \mathcal{L}(y) = \ln\gamma\left(\dfrac{\nu}{2}, \dfrac{y}{2}\right)

    where, :math:`\gamma\left(\cdot, \cdot\right)` is the :obj:`~ssp.gammainc` lower regularized incomplete
    gamma function, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}

    The final log CDF is expressed as :math:`\mathcal{L}(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    return np.where(y > 0, LOG(ssp.gammainc(degree_of_freedom / 2, y / 2)), -INF)


@suppress_numpy_warnings()
def cubic(
    x: ListOrNdArray,
    a: float = 1.0,
    b: float = 1.0,
    c: float = 1.0,
    d: float = 1.0,
) -> np.ndarray:
    r"""
    Computes the y-values of a cubic function given x-values.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    a : float
        The coefficient of the cubic term (x^3).
    b : float
        The coefficient of the quadratic term (x^2).
    c : float
        The coefficient of the linear term (x).
    d : float
        The constant term (y-intercept).

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The cubic function is defined as:

    .. math:: y = ax^3 + bx^2 + cx + d

    where, :math:`a`, math:`b`, :math:`c`, and :math:`d` are the cubic coefficients.
    """
    return a * x**3 + b * x**2 + c * x + d


@suppress_numpy_warnings()
def exponential_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    lambda_: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF for :class:`~pymultifit.distributions.exponential_d.ExponentialDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    lambda_ : float, optional
        The scale parameter, :math:`\lambda`.
        Defaults to 1.0.
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The Exponential PDF is defined as:

    .. math::
        f(y, \lambda) =
        \begin{cases}
        \exp(-y) &;& y \geq 0, \\
        0 &;& y < 0.
        \end{cases}

    where, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\theta}

    and :math:`\theta = \dfrac{1}{\lambda}`. The final PDF is expressed as :math:`f(y)/\theta`.
    """
    rate = 1 / lambda_
    y = preprocess_input(x=x, loc=loc, scale=rate)

    if y.size == 0:
        return y

    pdf_ = np.where(y >= 0, np.exp(-y), 0)
    pdf_ /= rate

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=exponential_pdf_, style=doc_style)
def exponential_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    lambda_: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF for :class:`~pymultifit.distributions.exponential_d.ExponentialDistribution`.

    Parameters
    ----------
    amplitude : float, optional
        For API consistency only.
    normalize : bool, optional
        For API consistency only.

    Notes
    -----
    The Exponential log PDF is defined as:

    .. math::
        \ell(y, \lambda) =
        \begin{cases}
        - y &;& y \geq 0, \\
        -\infty &;& y < 0.
        \end{cases}

    where, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\theta}

    and :math:`\theta = \dfrac{1}{\lambda}`. The final log PDF is expressed as :math:`\ell(y) - \ln(\theta)`.
    """
    rate = 1 / lambda_
    y = preprocess_input(x=x, loc=loc, scale=rate)

    if y.size == 0:
        return y

    log_pdf_ = np.where(y >= 0, -y, -INF)
    log_pdf_ -= LOG(rate)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=exponential_pdf_, style=doc_style)
def exponential_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    lambda_: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF of :class:`~pymultifit.distributions.exponential_d.ExponentialDistribution`.

    .. note::
        This function uses :obj:`ssp.gammainc` to calculate the CDF with
        :math:`a = 1` and :math:`x = \dfrac{x - \text{loc}}{\theta}`, where :math:`\theta = \dfrac{1}{\lambda}`.

    Parameters
    ----------
    amplitude : float, optional
        For API consistency only.
    normalize : bool, optional
        For API consistency only.

    Notes
    -----
    The Exponential CDF is defined as:

    .. math:: F(y) = 1 - \exp\left[-y\right].

    where, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\theta}

    and :math:`\theta = \dfrac{1}{\lambda}`. The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=1 / lambda_)

    if y.size == 0:
        return y

    return np.where(y >= 0, -ssp.expm1(-y), 0)


@suppress_numpy_warnings()
@doc_inherit(parent=exponential_cdf_, style=doc_style)
def exponential_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    lambda_: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF of :class:`~pymultifit.distributions.exponential_d.ExponentialDistribution`.

    .. note::
        This function uses log transformation of :obj:`ssp.gammainc` to calculate the log CDF with
        :math:`a = 1` and :math:`x = \dfrac{x - \text{loc}}{\theta}`, where :math:`\theta = \dfrac{1}{\lambda}`.

    Notes
    -----
    The Exponential log CDF is defined as:

    .. math:: \mathcal{L}(y) = \ln\left(1 -\exp(y)\right).

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\theta}.

    and :math:`\theta = \dfrac{1}{\lambda}`. The final log CDF is expressed as :math:`\mathcal{L}(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=1 / lambda_)

    if y.size == 0:
        return y

    return np.where(y >= 0, LOG(-ssp.expm1(-y)), -INF)


@suppress_numpy_warnings()
def folded_normal_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF for :class:`~pymultifit.distributions.foldedNormal_d.FoldedNormalDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    mean : float, optional
        The mean parameter, :math:`\mu`.
        Defaults to 0.0.
    sigma : float, optional
        The standard deviation parameter, :math:`\sigma`.
        Defaults to 1.0.
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The FoldedNormal PDF is defined as:

    .. math:: f(y\ |\ \mu, \sigma) = \phi(y\ |\ \mu, 1) + \phi(y\ |\ -\mu, 1),

    where :math:`\phi` is the PDF of :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution`,
    and :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}

    The final PDF is expressed as :math:`f(y)/\text{scale}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    pdf_ = np.where(
        y >= 0, gaussian_pdf_(y, mean=mean, normalize=True) + gaussian_pdf_(y, mean=-mean, normalize=True), 0
    )
    pdf_ /= sigma

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=folded_normal_pdf_, style=doc_style)
def folded_normal_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
):
    r"""
    Compute log PDF for :class:`~pymultifit.distributions.foldedNormal_d.FoldedNormalDistribution`.

    Notes
    -----
    The FoldedNormal PDF is defined as:

    .. math:: \ell(y\ |\ \mu, \sigma) = \ln\left(\phi(y\ |\ \mu, 1) + \phi(y\ |\ -\mu, 1)\right),

    where :math:`\phi` is the PDF of :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution`,
    and :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\sigma}

    The final log PDF is expressed as :math:`\ell(y) - \ln(\text{scale})`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    log_pdf_ = np.where(
        y >= 0, LOG(gaussian_pdf_(y, mean=mean, normalize=True) + gaussian_pdf_(y, mean=-mean, normalize=True)), -INF
    )
    log_pdf_ -= LOG(sigma)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=folded_normal_pdf_, style=doc_style)
def folded_normal_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF for :class:`~pymultifit.distributions.foldedNormal_d.FoldedNormalDistribution`.

    Parameters
    ----------
    amplitude : float, optional
        For API consistency only.
    normalize : bool, optional
        For API consistency only.

    Notes
    -----
    The FoldedNormal CDF is defined as:

    .. math:: F(y) = \Phi(y\ | \mu, 1) + \Phi(y\ | -\mu, 1) - 1

    where :math:`\Phi` is the CDF of :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution`,
    and :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\sigma}.

    The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    q = (y + mean) / SQRT_TWO
    r = (y - mean) / SQRT_TWO

    return np.where(y >= 0, _folded_cdf(q=q, r=r), 0)


@suppress_numpy_warnings()
@doc_inherit(parent=folded_normal_cdf_, style=doc_style)
def folded_normal_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF for :class:`~pymultifit.distributions.foldedNormal_d.FoldedNormalDistribution`.

    Notes
    -----
    The FoldedNormal log CDF is defined as:

    .. math:: \mathcal{L}(y) = -\ln(2) + \ln\left[\text{erf}\left(\dfrac{q}{\sqrt{2}}\right) +
              \text{erf}\left(\dfrac{r}{\sqrt{2}}\right)\right]

    where :math:`q = y + \mu`, :math:`r = y - \mu`, :math:`\text{erf}` is :obj:`~ssp.erf` function and
    :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\sigma}.

    The final logCDF is expressed as :math:`\mathcal{L}(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    q = (y + mean) / SQRT_TWO
    r = (y - mean) / SQRT_TWO

    return np.where(y >= 0, LOG(_folded_cdf(q=q, r=r)), -INF)


@suppress_numpy_warnings()
def _folded(
    x: ListOrNdArray,
    mean: float,
    loc: float,
    scale: float,
    g_func: Callable,
):
    r"""
    Precompute the gaussian part of :class:`~pymultifit.distributions.foldedNormal_d.FoldedNormalDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    mean : float, optional
        The mean parameter, :math:`\mu`.
        Defaults to 0.0.
    scale : float, optional
        The standard deviation parameter, :math:`\sigma`.
        Defaults to 1.0.
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    g_func : Callable
        The gaussian function, either PDF or CDF.

    Returns
    -------
    np.ndarray
        The additive gaussian part of the folded normal distribution.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    if scale <= 0 or mean < 0:
        return np.full(shape=x.size, fill_value=np.nan)

    g1 = g_func(x=y, mean=mean, normalize=True)
    g2 = g_func(x=y, mean=-mean, normalize=True)

    return y >= 0, np.where(y >= 0, g1 + g2, 0)


@suppress_numpy_warnings()
def gamma_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    theta: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF for :class:`~pymultifit.distributions.gamma_d.GammaDistribution`

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    alpha : float, optional
        The shape parameter, :math:`\alpha`.
        Defaults to 1.0.
    theta : float, optional
        The scale parameter, :math:`\theta`.
        Defaults to 1.0.
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    .. important::

        The Gamma SS PDF is calculated via exponentiation of :func:`gamma_sr_log_pdf_` by setting
        :math:`\lambda = \dfrac{1}{\theta}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=theta)

    if y.size == 0:
        return y

    pdf_ = _gamma(x=y, a=alpha, un_log=True)
    pdf_ /= theta

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=gamma_pdf_, style=doc_style)
def gamma_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    theta: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF for :class:`~pymultifit.distributions.gamma_d.GammaDistribution`

    Notes
    -----
    .. important::

        The Gamma SS log PDF is calculated via :func:`gamma_sr_log_pdf_` by setting :math:`\lambda = \dfrac{1}{\theta}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=theta)

    if y.size == 0:
        return y

    log_pdf_ = _gamma(x=y, a=alpha)
    log_pdf_ -= LOG(theta)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


def _gamma(x, a, un_log=False):
    value = np.where(x >= 0, ssp.xlogy(a - 1.0, x) - x - ssp.gammaln(a), -INF)
    return np.exp(value) if un_log else value


@suppress_numpy_warnings()
@doc_inherit(parent=gamma_pdf_, style=doc_style)
def gamma_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    theta: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF for :class:`~pymultifit.distributions.gamma_d.GammaDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: bool, optional
        For API consistency only.

    Notes
    -----
    .. important::

        The Gamma SS CDF is calculated via :func:`gamma_sr_cdf_` by setting :math:`\lambda = \dfrac{1}{\theta}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=theta)

    if y.size == 0:
        return y

    return np.where(y > 0, ssp.gammainc(alpha, y), 0)


@suppress_numpy_warnings()
@doc_inherit(parent=gamma_cdf_, style=doc_style)
def gamma_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    alpha: float = 1.0,
    theta: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF for :class:`~pymultifit.distributions.gamma_d.GammaDistribution`.

    Notes
    -----
    .. important::

        The Gamma SS log CDF is calculated via logarithm of :func:`gamma_sr_cdf_` by setting
        :math:`\lambda = \dfrac{1}{\theta}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=theta)

    if y.size == 0:
        return y

    return LOG(np.where(y > 0, ssp.gammainc(alpha, y), 0))


@suppress_numpy_warnings()
def gaussian_pdf_(
    x: ListOrNdArray,
    amplitude=1.0,
    mean=0.0,
    std=1.0,
    normalize=False,
) -> np.ndarray:
    r"""
    Compute PDF for :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution`

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    mean : float, optional
        The mean parameter, :math:`\mu`.
        Defaults to 0.0.
    std : float, optional
        The standard deviation parameter, :math:`\sigma`.
        Defaults to 1.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The Gaussian PDF is defined as:

    .. math::
        f(x; \mu, \sigma) = \phi\left(\dfrac{x-\mu}{\sigma}\right) =
        \dfrac{1}{\sqrt{2\pi\sigma}}\exp\left[-\dfrac{1}{2}\left(\dfrac{x-\mu}{\sigma}\right)^2\right]

    The final PDF is expressed as :math:`f(x)`.
    """
    y = preprocess_input(x=x, loc=mean, scale=std)

    if y.size == 0:
        return y

    pdf_ = np.exp(-0.5 * y**2) / SQRT_TWO_PI / std

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=gaussian_pdf_, style=doc_style)
def gaussian_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    std: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF for :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution`

    Notes
    -----
    The Gaussian log PDF is defined as:

    .. math::
        \ell(x; \mu, \sigma) = -\dfrac{1}{2}\ln(2\pi) - \ln\sigma - \dfrac{1}{2}\left(\dfrac{x-\mu}{\sigma}\right)^2

    The final log PDF is expressed as :math:`\ell(x)`.
    """
    y = preprocess_input(x=x, loc=mean, scale=std)

    if y.size == 0:
        return y

    log_pdf_ = -(y**2) / 2.0 - LOG_SQRT_TWO_PI - LOG(std)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=gaussian_pdf_, style=doc_style)
def gaussian_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    std: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF for :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution`

    .. important::

        The calculation of gaussian CDF is done using :obj:`ssp.ndtr` function.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: float, optional
        For API consistency only.

    Notes
    -----
    The Gaussian CDF is defined as:

    .. math::
        F(x) = \Phi\left(\dfrac{x-\mu}{\sigma}\right) =
        \dfrac{1}{2} \left[1 + \text{erf}\left(\dfrac{x - \mu}{\sigma\sqrt{2}}\right)\right]

    The final CDF is expressed as :math:`F(x)`.
    """
    return ssp.ndtr((x - mean) / std)


@suppress_numpy_warnings()
@doc_inherit(parent=gaussian_cdf_, style=doc_style)
def gaussian_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    std: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF for :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution`

    .. important::

        The calculation of gaussian log CDF is done using :obj:`ssp.log_ndtr` function.

    Notes
    -----
    The Gaussian log CDF is defined as:

    .. math::
        \mathcal{L}(x) = \ln\Phi\left(\dfrac{x-\mu}{\sigma}\right)

    The final log CDF is expressed as :math:`\mathcal{L}(x)`.
    """
    return ssp.log_ndtr((x - mean) / std)


@suppress_numpy_warnings()
def half_normal_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF for the :class:`~pymultifit.distributions.halfNormal_d.HalfNormalDistribution`.

    .. note::
        The :class:`~pymultifit.distributions.halfNormal_d.HalfNormalDistribution` is a special case of the
        :class:`~pymultifit.distributions.foldedNormal_d.FoldedNormalDistribution` with :math:`\mu = 0`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    sigma : float, optional
        The standard deviation :math:`\sigma`.
        Defaults to 1.0.
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The HalfNormal PDF is defined as:

    .. math::
        f(y\ |\ \sigma) = \sqrt{\dfrac{2}{\pi}}\exp\left(-\dfrac{y^2}{2}\right)

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}.

    The final PDF is expressed as :math:`f(y)/\text{scale}`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    pdf_ = np.where(y >= 0, SQRT_TWO_BY_PI * np.exp(-0.5 * y**2), 0)
    pdf_ /= sigma

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=half_normal_pdf_, style=doc_style)
def half_normal_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF for the :class:`~pymultifit.distributions.halfNormal_d.HalfNormalDistribution`.

    Notes
    -----
    The HalfNormal log PDF is defined as:

    .. math:: \ell(y\ |\ \sigma) = \dfrac{1}{2}\ln\left(\dfrac{2}{\pi}\right) - \dfrac{y^2}{2}

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}.

    The final log PDF is expressed as :math:`\ell(y) - \ln\left(\text{scale}\right)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    log_pdf_ = np.where(y >= 0, LOG_SQRT_TWO_BY_PI - 0.5 * y**2, -INF)
    log_pdf_ -= LOG(sigma)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=half_normal_pdf_, style=doc_style)
def half_normal_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute the CDF for :class:`~pymultifit.distributions.halfNormal_d.HalfNormalDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: float, optional
        For API consistency only.

    Notes
    -----
    The HalfNormal CDF is defined as:

    .. math:: F(y) = \text{erf}\left(\frac{y}{\sqrt{2}}\right)

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}.

    The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    return np.where(y >= 0, ssp.erf(y / SQRT_TWO), 0)


@suppress_numpy_warnings()
@doc_inherit(parent=half_normal_cdf_, style=doc_style)
def half_normal_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    sigma: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute the log CDF for :class:`~pymultifit.distributions.halfNormal_d.HalfNormalDistribution`.

    Notes
    -----
    The HalfNormal log CDF is defined as:

    .. math:: \mathcal{L}(y) = \ln\text{erf}\left(\frac{y}{\sqrt{2}}\right)

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \text{loc}}{\text{scale}}.

    The final log CDF is expressed as :math:`\mathcal{L}(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=sigma)

    if y.size == 0:
        return y

    return np.where(y >= 0, LOG(ssp.erf(y / SQRT_TWO)), -INF)


@suppress_numpy_warnings()
def laplace_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    diversity: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF for the :class:`~pymultifit.distributions.laplace_d.LaplaceDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF. Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    mean : float, optional
        The mean of laplace distribution.
        Defaults to 0.0.
    diversity : float, optional
        The diversity parameter for laplace distribution.
        Defaults to 1.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The Laplace PDF is defined as:

    .. math:: f(y\ |\ \mu, b) = \dfrac{1}{2b}\exp\left(-\dfrac{|y|}{b}\right)

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = x - \mu.

    The final PDF is expressed as :math:`f(y)`.
    """
    y = preprocess_input(x=x, loc=mean, scale=diversity)

    if y.size == 0:
        return y

    pdf_ = (1 / 2) * np.exp(-np.abs(y))
    pdf_ /= diversity

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=laplace_pdf_, style=doc_style)
def laplace_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    diversity: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF for the :class:`~pymultifit.distributions.laplace_d.LaplaceDistribution`.

    Notes
    -----
    The Laplace log PDF is defined as:

    .. math:: \ell(y\ |\ \mu, b) = -\ln(2b) - \dfrac{|y|}{b}

    where :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = x - \mu

    The final log PDF is expressed as :math:`\ell(y)`.
    """
    y = preprocess_input(x=x, loc=mean, scale=diversity)

    if y.size == 0:
        return y

    log_pdf_ = LOG(0.5 * np.exp(-np.abs(y)))
    log_pdf_ -= LOG(diversity)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=laplace_pdf_, style=doc_style)
def laplace_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    diversity: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF for :class:`~pymultifit.distributions.laplace_d.LaplaceDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: bool, optional
        For API consistency only.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The Laplace CDF is defined as:

    .. math:: F(x) =
        \begin{cases}
        \dfrac{1}{2}\exp\left(\dfrac{x-\mu}{b}\right) &,&x\leq\mu\\
        1 - \dfrac{1}{2}\exp\left(-\dfrac{x-\mu}{b}\right) &,&x\geq\mu
        \end{cases}

    The final CDF is expressed as :math:`F(x)`.
    """
    y = preprocess_input(x=x, loc=mean, scale=diversity)

    if y.size == 0:
        return y

    return np.where(y > 0, 1.0 - 0.5 * np.exp(-y), 0.5 * np.exp(y))


@suppress_numpy_warnings()
@doc_inherit(parent=laplace_cdf_, style=doc_style)
def laplace_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 0.0,
    diversity: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF for :class:`~pymultifit.distributions.laplace_d.LaplaceDistribution`.

    Notes
    -----
    The Laplace log CDF is defined as:

    .. math:: \mathcal{L}(x) =
        \begin{cases}
        -\ln(2) + \dfrac{x-\mu}{b} &,&x\leq\mu\\
        \ln\left[1 - \dfrac{1}{2}\exp\left(-\dfrac{x-\mu}{b}\right)\right] &,&x\geq\mu
        \end{cases}
    """
    y = preprocess_input(x=x, loc=mean, scale=diversity)

    if y.size == 0:
        return y

    return np.where(y > 0, np.log1p(-0.5 * np.exp(-y)), -LOG_TWO + y)


@suppress_numpy_warnings()
def line(
    x: ListOrNdArray,
    slope: float = 1.0,
    intercept: float = 0.0,
) -> np.ndarray:
    r"""
    Computes the y-values of a line given x-values, slope, and intercept.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    slope : float
        The slope of the line.
    intercept : float
        The y-intercept of the line.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The line/linear function is defined as:

    .. math:: y = mx + c

    where :math:`m` is the slope and :math:`c` is the intercept of the function.
    """
    return slope * x + intercept


@suppress_numpy_warnings()
def log_normal_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 1.0,
    std: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF for :class:`~pymultifit.distributions.logNormal_d.LogNormalDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF.
        Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    mean : float, optional
        The mean parameter, :math:`\mu`.
        Defaults to 0.0.
    std : float, optional
        The standard deviation parameter, :math:`\sigma`.
        Defaults to 1.0.
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The LogNormal PDF is defined as:

    .. math::
        f(y\ |\ \mu, \sigma) = \dfrac{1}{\sigma y\sqrt{2\pi}}\exp\left(-\dfrac{(\ln y - \mu)^2}{2\sigma^2}\right)

    where, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = x - \text{loc}

    The final PDF is expressed as :math:`f(y)`.
    """
    y = preprocess_input(x=x, loc=loc)

    if y.size == 0:
        return y

    q = (LOG(y) - mean) / std

    pdf_ = np.where(y > 0, 1 / y / np.exp(q**2 / 2) / SQRT_TWO_PI, 0)
    pdf_ /= std

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=log_normal_pdf_, style=doc_style)
def log_normal_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 1.0,
    std: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF for :class:`~pymultifit.distributions.logNormal_d.LogNormalDistribution`.

    Notes
    -----
    The LogNormal log PDF is defined as:

    .. math::
        f(y\ |\ \mu, \sigma) = -\ln(\sigma) -\ln(y) - 0.5\ln(2\pi) -\dfrac{1}{2}\dfrac{(\ln y - \mu)^2}{\sigma^2}

    where, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = x - \text{loc}

    The final PDF is expressed as :math:`f(y)`.
    """
    y = preprocess_input(x=x, loc=loc)

    if y.size == 0:
        return y

    q = (LOG(y) - mean) / std

    log_pdf_ = np.where(y > 0, -LOG(y) - (q**2 / 2.0) - LOG_SQRT_TWO_PI, -INF)
    log_pdf_ -= LOG(std)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=log_normal_pdf_, style=doc_style)
def log_normal_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 1.0,
    std=1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF of :class:`~pymultifit.distributions.logNormal_d.LogNormalDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: bool, optional
        For API consistency only

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    .. important::
        The LogNormal CDF is defined as:

        .. math::
            F(x) = \Phi\left(\dfrac{\ln x - \mu}{\sigma}\right)

        which can be calculated via :obj:`ssp.ndtr` function with ``ndtr(y)``, where :math:`y` is the
        transformed value of :math:`x`, defined as:

        .. math:: y = \dfrac{\ln(x - \text{loc}) - \mu}{\sigma}.
    """
    y = preprocess_input(x=x, loc=loc, scale=np.exp(mean))

    return np.where(y > 0, ssp.ndtr(LOG(y) / std), 0)


@suppress_numpy_warnings()
@doc_inherit(parent=log_normal_cdf_, style=doc_style)
def log_normal_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    mean: float = 1.0,
    std: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF of :class:`~pymultifit.distributions.logNormal_d.LogNormalDistribution`.

    Notes
    -----
    .. important::
        The LogNormal log CDF is defined as:

        .. math::
            F(x) = \ln\left[\Phi\left(\dfrac{\ln x - \mu}{\sigma}\right)\right]

        which can be calculated via :obj:`ssp.log_ndtr` function function with ``log_ndtr(y)``, where :math:`y`
        is the transformed value of :math:`x`, defined as:

        .. math:: y = \dfrac{\ln(x - \text{loc}) - \mu}{\sigma}.
    """
    y = preprocess_input(x=x, loc=loc, scale=np.exp(mean))

    return np.where(y > 0, ssp.log_ndtr(LOG(y) / std), -INF)


@suppress_numpy_warnings()
def uniform_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    low: float = 0.0,
    high: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF of :class:`~pymultifit.distributions.uniform_d.UniformDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF.
        Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    low : float, optional
        The lower bound, :math:`a`.
        Defaults to 0.0.
    high : float, optional
        The upper bound, :math:`b`.
        Defaults to 1.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The Uniform PDF is defined as:

    .. math:: f(x\ |\ a, b) = \dfrac{1}{\beta - a}

    Where :math:`\beta = a + b` consistent with ``loc`` and ``scale`` factors and the final PDF is expressed as,
    :math:`f(x)`.
    """
    x = np.asarray(a=x, dtype=float)

    if x.size == 0:
        return x

    high_ = high + low

    if high_ == low:
        return np.full(shape=x.size, fill_value=np.nan)

    pdf_ = np.where((x >= low) & (x <= high_), 1 / high, 0)

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
def uniform_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    low: float = 0.0,
    high: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF of :class:`~pymultifit.distributions.uniform_d.UniformDistribution`.

    Notes
    -----
    The Uniform log PDF is defined as:

    .. math:: \ell(x\ |\ a, b) = -\ln(\beta - a)

    where :math:`\beta = a + b` is consistent with ``loc`` and ``scale`` factors, and the final logPDF is expressed as,
    :math:`\ell(x)`.
    """
    x = np.asarray(a=x, dtype=float)

    if x.size == 0:
        return x

    high_ = high + low

    log_pdf_ = np.where((x >= low) & (x <= high_), -LOG(high), -INF)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=uniform_pdf_, style=doc_style)
def uniform_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    low: float = 0.0,
    high: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF of :class:`~pymultifit.distributions.uniform_d.UniformDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: bool, optional
        For API consistency only.

    Notes
    -----
    The Uniform CDF is defined as:

    .. math:: F(x) = \begin{cases}
                        0 &,& x < a\\
                        \dfrac{x-a}{b-a} &,& x \in [a, b]\\
                        1 &,& x > b
                        \end{cases}
    """
    y = preprocess_input(x=x, loc=low, scale=high)

    if y.size == 0:
        return y

    high_ = high + low

    if low == high_ == 0:
        return np.full(shape=y.size, fill_value=np.nan)

    return np.select(condlist=[y < 0, y > 1], choicelist=[0, 1], default=y)


@suppress_numpy_warnings()
@doc_inherit(parent=uniform_cdf_, style=doc_style)
def uniform_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    low: float = 0.0,
    high: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF of :class:`~pymultifit.distributions.uniform_d.UniformDistribution`.

    Notes
    -----
    The Uniform log CDF is defined as:

    .. math:: \mathcal{L}(x) = \begin{cases}
                        -\infty &,& x < a\\
                        \ln\left(\dfrac{x-a}{\beta-a}\right) &,& x \in [a, b]\\
                        0 &,& x > \beta
                        \end{cases}

    The final logCDF is expressed as, :math:`\mathcal{L}(x)`.
    """
    y = preprocess_input(x=x, loc=low, scale=high)

    if y.size == 0:
        return y

    high_ = high + low

    if low == high_ == 0:
        return np.full(shape=y.shape, fill_value=-INF)

    return np.select(condlist=[y < 0, y > 1], choicelist=[-INF, 0], default=LOG(y))


@suppress_numpy_warnings()
def scaled_inv_chi_square_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    df: float = 1.0,
    scale: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
):
    r"""
    Compute PDF of :class:`~pymultifit.distributions.scaledInvChiSquare_d.ScaledInverseChiSquareDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF.
        Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    df : float, optional
        The degree of freedom.
        Defaults to 1.0.
    scale: float, optional
        The scale parameter, for scaling.
        Defaults to 1.0,
    loc : float, optional
        The location parameter, for shifting.
        Defaults to 0.0.
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The Scaled Inverse ChiSquare PDF is defined as:

    .. math:: f(y\ | \nu,\phi) = \dfrac{\tau^2\nu_2}{\Gamma(\nu_2)}\dfrac{1}{y^{1+\nu_2}}\exp\left[-\dfrac{\nu\tau^2}{2y}\right]

    where :math:`\nu_2 = \dfrac{\nu}{2}`, :math:`\tau^2 = \dfrac{\phi}{\nu}` and :math:`y` is the transformed
    value of :math:`x`, defined as:

    .. math:: y = x - \text{loc}

    The final PDF is expressed as :math:`f(y)`.
    """
    y = preprocess_input(x=x, loc=loc)

    if y.size == 0:
        return y

    tau2 = scale / df
    df_half = df / 2

    f1 = np.power(tau2 * df_half, df_half) * ssp.rgamma(df_half)
    f2 = np.exp(-(df * tau2) / (2 * y)) / np.power(y, 1 + df_half)

    pdf_ = np.where(y > 0, f1 * f2, 0)

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=scaled_inv_chi_square_pdf_, style=doc_style)
def scaled_inv_chi_square_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    df: Union[int, float] = 1.0,
    scale: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
):
    r"""
    Compute logPDF of :class:`~pymultifit.distributions.scaledInvChiSquare_d.ScaledInverseChiSquareDistribution`.

    Notes
    -----
    The Scaled Inverse ChiSquare PDF is defined as:

    .. math:: \ell(y) = \ln(\tau^2\nu_2) - \ln\Gamma(\nu_2) - (1+\nu_2)\ln(\nu) - \dfrac{\nu\tau^2}{2y}

    where :math:`\ln` is the natural logarithm, :math:`\ln\Gamma(\cdot)` is the :obj:`~ssp.gammaln` function,
    :math:`\nu_2 = \dfrac{\nu}{2}`, :math:`\tau^2 = \dfrac{\phi}{\nu}` and :math:`y` is the transformed value of
    :math:`x`, defined as:

    .. math:: y = x - \text{loc}

    The final PDF is expressed as :math:`\ell(y)`.
    """
    y = preprocess_input(x=x, loc=loc)

    if y.size == 0:
        return y

    tau2 = scale / df
    df_half = df / 2

    f1 = ssp.xlogy(df_half, tau2 * df_half) - ssp.gammaln(df_half)
    f2 = -(tau2 * df) / (2 * y) - ssp.xlogy(1 + df_half, y)

    log_pdf_ = np.where(y > 0, f1 + f2, -INF)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=scaled_inv_chi_square_pdf_, style=doc_style)
def scaled_inv_chi_square_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    df: Union[int, float] = 1.0,
    scale: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
):
    r"""
    Compute CDF of :class:`~pymultifit.distributions.scaledInvChiSquare_d.ScaledInverseChiSquareDistribution`.

    Parameters
    ----------
    amplitude : float, optional
        For API consistency only.
    normalize : bool, optional
        For API consistency only.

    Notes
    -----
    The Scaled Inverse ChiSquare CDF is defined as:

    .. math:: F(y) = \Gamma\left(\nu_2, \dfrac{\tau^2\nu_2}{y}\right)

    where :math:`\nu_2 = \dfrac{\nu}{2}`, :math:`\tau^2 = \dfrac{\phi}{\nu}`, :math:`\Gamma(a, b)` is the regularized
    upper gamma function, see :obj:`ssp.gammaincc`,and :math:`y` is the transformed value of :math:`x`,
    defined as:

    .. math:: y = x - \text{loc}

    The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc)

    if y.size == 0:
        return y

    tau2 = scale / df
    df_half = df / 2

    return np.where(y > 0, ssp.gammaincc(df_half, (tau2 * df_half) / y), 0)


@suppress_numpy_warnings()
@doc_inherit(parent=scaled_inv_chi_square_pdf_, style=doc_style)
def scaled_inv_chi_square_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    df: Union[int, float] = 1.0,
    scale: float = 1.0,
    loc: float = 0.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF of :class:`~pymultifit.distributions.scaledInvChiSquare_d.ScaledInverseChiSquareDistribution`.

    Notes
    -----
    The Scaled Inverse ChiSquare log CDF is defined as:

    .. math:: \mathcal{L}(y) = \ln\left[\Gamma\left(\nu_2, \dfrac{\tau^2\nu_2}{y}\right)\right]

    where :math:`\nu_2 = \dfrac{\nu}{2}`, :math:`\tau^2 = \dfrac{\phi}{\nu}`, :math:`\Gamma(a, b)` is the regularized
    upper gamma function, see :obj:`ssp.gammaincc`,and :math:`y` is the transformed value of :math:`x`,
    defined as:

    .. math:: y = x - \text{loc}

    The final log CDF is expressed as :math:`\mathcal{L}(y)`.
    """
    y = preprocess_input(x=x, loc=loc)

    if y.size == 0:
        return y

    tau2 = scale / df
    df_half = df / 2

    return np.where(y > 0, LOG(ssp.gammaincc(df_half, (tau2 * df_half) / y)), -INF)


@suppress_numpy_warnings()
def skew_normal_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    shape: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF of :class:`~pymultifit.distributions.skewNormal_d.SkewNormalDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF.
        Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    shape : float, optional
        The shape parameter, :math:`\alpha`.
        Defaults to 0.0.
    loc : float, optional
        The location parameter, :math:`\xi`.
        Defaults to 0.0.
    scale: float, optional
        The scale parameter, :math:`\omega`
        Defaults to 1.0,
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The SkewNormal PDF is defined as:

    .. math:: f(y\ |\ \alpha, \xi, \omega) =
             2\phi(y)\Phi(\alpha y)

    where, :math:`\phi(y)` and :math:`\Phi(\alpha y)` are the
    :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution` PDF and CDF defined at :math:`y` and
    :math:`\alpha y` respectively. Additionally, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \xi}{\omega}

    The final PDF is expressed as :math:`f(y)/\omega`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    pdf_ = 2 * gaussian_pdf_(x=y, normalize=True) * gaussian_cdf_(x=shape * y, normalize=True)
    pdf_ /= scale

    if not normalize:
        pdf_ = _pdf_scaling(pdf_=pdf_, amplitude=amplitude)

    return pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=skew_normal_pdf_, style=doc_style)
def skew_normal_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    shape: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF of :class:`~pymultifit.distributions.skewNormal_d.SkewNormalDistribution`.

    Notes
    -----
    The SkewNormal log PDF is defined as:

    .. math:: \ell(y\ |\ \alpha, \xi, \omega) = \ln(2) + \ln\phi(y) + \ln\Phi(\alpha y)

    where, :math:`\phi(y)` and :math:`\Phi(\alpha y)` are the
    :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution` PDF and CDF defined at :math:`y` and
    :math:`\alpha y` respectively. Additionally, :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \dfrac{x - \xi}{\omega}

    The final log PDF is expressed as :math:`\ell(y)/\omega`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    log_pdf_ = LOG_TWO + gaussian_log_pdf_(x=y, normalize=True) + gaussian_log_cdf_(x=shape * y, normalize=True)
    log_pdf_ -= LOG(scale)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=skew_normal_pdf_, style=doc_style)
def skew_normal_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    shape: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
):
    r"""
    Compute CDF of :class:`~pymultifit.distributions.skewNormal_d.SkewNormalDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: float, optional
        For API consistency only.

    Notes
    ------
    The SkewNormal CDF is defined as:

    .. math:: F(y) = \Phi(y) - 2T(y, \alpha)

    where, :math:`T` is the Owen's T function, see :obj:`ssp.owens_t`, and
    :math:`\Phi(\cdot)` is the :class:`~pymultifit.distributions.gaussian_d.GaussianDistribution` CDF function, and
    :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    return gaussian_cdf_(x=y, normalize=True) - 2 * ssp.owens_t(y, shape)


@suppress_numpy_warnings()
def sym_gen_normal_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    shape: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute PDF of :class:`~pymultifit.distributions.generalized.genNorm_d.SymmetricGeneralizedNormalDistribution`.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    amplitude : float, optional
        The amplitude of the PDF.
        Defaults to 1.0.
        Ignored if **normalize** is ``True``.
    shape : float, optional
        The shape parameter, :math:`\beta`.
        Defaults to 1.0.
    loc : float, optional
        The location parameter, :math:`\mu`.
        Defaults to 0.0.
    scale: float, optional
        The scale parameter, :math:`\alpha`
        Defaults to 1.0,
    normalize : bool, optional
        If ``True``, the distribution is normalized so that the total area under the PDF equals 1.
        Defaults to ``False``.

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The SymmetricGeneralizedNormalDistribution PDF is defined as:

    .. math:: f(y\ |\ \beta, \mu, \alpha) = \dfrac{\beta}{2\Gamma(1/\beta)}\exp\left(-|y|^\beta\right)

    where, :math:`\Gamma` is the :obj:`ssp.gamma` function, and :math:`y` is the transformed value of
    :math:`x`, defined as:

    .. math:: y = \frac{x - \mu}{\alpha}

    The final PDF is expressed as :math:`f(y)/\alpha`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    _, _, beta = loc, scale, shape

    log_pdf_ = beta / 2 / ssp.gamma(1 / beta) * np.exp(-np.power(np.abs(y), beta))
    log_pdf_ /= scale

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(sym_gen_normal_pdf_, style=doc_style)
def sym_gen_normal_log_pdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    shape: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log PDF of :class:`~pymultifit.distributions.generalized.genNorm_d.SymmetricGeneralizedNormalDistribution`.

    Notes
    -----
    The SymmetricGeneralizedNormalDistribution log PDF is defined as:

    .. math:: \ell(y\ |\ \beta, \mu, \alpha) = \ln(\beta) - \ln(2) - \ln\Gamma\left(\dfrac{1}{\beta}\right) - |y|^\beta

    where, :math:`\Gamma` is the :obj:`ssp.gamma` function, and :math:`y` is the transformed value of
    :math:`x`, defined as:

    .. math:: y = \frac{x - \mu}{\alpha}

    The final log PDF is expressed as :math:`\ell(y)/\alpha`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    _, _, beta = loc, scale, shape

    log_pdf_ = LOG(beta) - LOG_TWO - ssp.gammaln(1 / beta) - np.power(np.abs(y), beta)
    log_pdf_ -= LOG(scale)

    if not normalize:
        log_pdf_ = _log_pdf_scaling(log_pdf_=log_pdf_, amplitude=amplitude)

    return log_pdf_


@suppress_numpy_warnings()
@doc_inherit(parent=sym_gen_normal_pdf_, style=doc_style)
def sym_gen_normal_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    shape: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute CDF of :class:`~pymultifit.distributions.generalized.genNorm_d.SymmetricGeneralizedNormalDistribution`.

    Parameters
    ----------
    amplitude: float, optional
        For API consistency only.
    normalize: bool, optional
        For API consistency only.

    Notes
    -----
    The SymmetricGeneralizedNormalDistribution CDF is defined as:

    .. math:: F(y) = \dfrac{1}{2} + \dfrac{\text{sign}(y)}{2}\gamma\left(\dfrac{1}{\beta},|y|^\beta\,\right)

    where :math:`\gamma(\cdot,\cdot)` is the regularized lower incomplete gamma function, see
    :obj:`~ssp.gammainc`, and :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final CDF is expressed as :math:`F(y)`.
    """
    y = preprocess_input(x=x, loc=loc, scale=scale)

    if y.size == 0:
        return y

    _, _, beta = loc, scale, shape

    return 0.5 + np.sign(y) * 0.5 * ssp.gammainc(1 / beta, np.power(np.abs(y), beta))


@suppress_numpy_warnings()
@doc_inherit(parent=sym_gen_normal_cdf_, style=doc_style)
def sym_gen_normal_log_cdf_(
    x: ListOrNdArray,
    amplitude: float = 1.0,
    shape: float = 1.0,
    loc: float = 0.0,
    scale: float = 1.0,
    normalize: bool = False,
) -> np.ndarray:
    r"""
    Compute log CDF of :class:`~pymultifit.distributions.generalized.genNorm_d.SymmetricGeneralizedNormalDistribution`.

    Notes
    -----
    The SymmetricGeneralizedNormalDistribution log CDF is defined as:

    .. math:: \mathcal{L}(y) =
     \ln\left[\dfrac{1}{2} + \dfrac{\text{sign}(y)}{2}\gamma\left(\dfrac{1}{\beta},|y|^\beta\,\right)\right]

    where :math:`\gamma(\cdot,\cdot)` is the lower incomplete gamma function, see :obj:`~ssp.gammainc`, and
    :math:`y` is the transformed value of :math:`x`, defined as:

    .. math:: y = \frac{x - \text{loc}}{\text{scale}}

    The final PDF is expressed as :math:`f(y)/\text{scale}`.

    """
    cdf_ = sym_gen_normal_cdf_(x=x, amplitude=amplitude, shape=shape, loc=loc, scale=scale, normalize=normalize)

    return LOG(cdf_)


@suppress_numpy_warnings()
def quadratic(
    x: ListOrNdArray,
    a: float = 1.0,
    b: float = 1.0,
    c: float = 1.0,
) -> np.ndarray:
    r"""
    Computes the y-values of a quadratic function given x-values.

    Parameters
    ----------
    x : np.ndarray
        Input array of values.
    a : float
        The coefficient of the quadratic term (x^2).
    b : float
        The coefficient of the linear term (x).
    c : float
        The constant term (y-intercept).

    Returns
    -------
    np.ndarray
        Array of the same shape as :math:`x`, containing the evaluated values.

    Notes
    -----
    The quadratic function is defined as:

    .. math:: y = ax^2 + bx + c

    where, :math:`a`, :math:`b`, and :math:`c` are the quadratic coefficients.
    """
    return a * x**2 + b * x + c


@suppress_numpy_warnings()
def _beta_expr(
    y: np.ndarray,
    a: float,
    b: float,
    un_log: bool = False,
):
    in_range = (y > 0) & (y < 1)

    undefined_0 = (y == 0) & (a <= 1)
    undefined_1 = (y == 1) & (b <= 1)
    special_case = (y == 1) & (a == 1) & (b == 1)

    expr = ssp.xlog1py(b - 1.0, -y) + ssp.xlogy(a - 1, y) - ssp.betaln(a, b)
    expr2 = np.power(y, a - 1) * np.power(1.0 - y, b - 1.0) / ssp.beta(a, b)

    return [special_case, undefined_0 | undefined_1, in_range], expr2 if un_log else expr


@suppress_numpy_warnings()
def _folded_cdf(
    q: float,
    r: float,
) -> float:
    return 0.5 * (ssp.erf(r) + ssp.erf(q))


@suppress_numpy_warnings()
def _pdf_scaling(
    pdf_: np.ndarray,
    amplitude: float,
) -> np.ndarray:
    """Scales a given PDF by a specified amplitude, normalizing it relative to its maximum value.

    Parameters
    ----------
    pdf_ : np.ndarray
        The input probability density function values (not necessarily normalized).
    amplitude : float
        The amplitude factor to scale the normalized PDF.

    Returns
    -------
    NdArray
        The scaled PDF array.
    """
    with np.errstate(all="ignore"):
        return amplitude * (pdf_ / np.max(pdf_))


@suppress_numpy_warnings()
def _log_pdf_scaling(
    log_pdf_: np.ndarray,
    amplitude: float,
) -> np.ndarray:
    with np.errstate(all="ignore"):
        return LOG(amplitude) + (log_pdf_ - np.max(log_pdf_))


@suppress_numpy_warnings()
def preprocess_input(
    x: ListOrNdArray,
    loc: float = 0.0,
    scale: float = 1.0,
) -> np.ndarray:
    """
    Preprocess the input array, checking for scalar input, handling empty arrays, and loc-scale normalizaing the data.

    Parameters
    ----------
    x : np.ndarray
        Input data.
    loc : float, optional
        The location parameter, for shifting, defaults to 0.0.
    scale: float, optional
        The scale parameter, for scaling, defaults to 1.0,

    Returns
    -------
    np.ndarray
        loc-scale shifted numpy array.
    """
    x = np.asarray(a=x, dtype=float)

    if x.size == 0:
        return np.array([])

    return (x - loc) / scale



"""Created on Aug 10 23:37:54 2024"""

import numpy as np

from .backend import BaseFitter
from .utilities_f import sanity_check
from .. import Params_
from ..distributions.utilities_d import line, quadratic, cubic


class LineFitter(BaseFitter):

    def __init__(self, x_values, y_values, max_iterations: int = 1000):
        x_values, y_values = sanity_check(x_values=x_values, y_values=y_values)
        super().__init__(x_values=x_values, y_values=y_values, max_iterations=max_iterations)
        self.n_par = 2

    def fit_boundaries(self):
        lb = (-np.inf, -np.inf)
        ub = (np.inf, np.inf)

        return lb, ub

    @staticmethod
    def fitter(x, params: Params_):
        return line(x, *params)


class QuadraticFitter(BaseFitter):
    def __init__(self, x_values, y_values, max_iterations: int = 1000):
        x_values, y_values = sanity_check(x_values=x_values, y_values=y_values)
        super().__init__(x_values=x_values, y_values=y_values, max_iterations=max_iterations)
        self.n_par = 3

    def fit_boundaries(self):
        lb = (-np.inf, -np.inf, -np.inf)
        ub = (np.inf, np.inf, np.inf)

        return lb, ub

    @staticmethod
    def fitter(x, params: Params_):
        return quadratic(x, *params)


class CubicFitter(BaseFitter):

    def __init__(self, x_values, y_values, max_iterations: int = 1000):
        x_values, y_values = sanity_check(x_values=x_values, y_values=y_values)
        super().__init__(x_values=x_values, y_values=y_values, max_iterations=max_iterations)
        self.n_par = 4

    def fit_boundaries(self):
        lb = (-np.inf, -np.inf, -np.inf, -np.inf)
        ub = (np.inf, np.inf, np.inf, np.inf)

        return lb, ub

    @staticmethod
    def fitter(x, params: Params_):
        return cubic(x, *params)



"""Created on Aug 03 22:06:29 2024"""

from typing import Dict, Optional

import numpy as np


class BaseDistribution:
    r"""
    Bare-bones class for statistical distributions to provide consistent methods.

    This class serves as a template for other distribution classes, defining the common interface
    for probability density function (PDF), cumulative distribution function (CDF), and statistics.
    """

    def pdf(self, x: np.ndarray) -> np.ndarray: # type: ignore[empty-body]
        r"""
        Compute the probability density function (PDF) for the distribution.

        :param x: Input array at which to evaluate the PDF.
        """
        raise NotImplementedError("Subclasses should implement this method.")

    def logpdf(self, x: np.ndarray) -> np.ndarray: # type: ignore[empty-body]
        r"""
        Compute the log probability density function (logPDF) for the distribution.

        :param x: Input array at which to evaluate the logPDF.
        """
        pass

    def cdf(self, x: np.ndarray) -> np.ndarray: # type: ignore[empty-body]
        """
        Compute the cumulative density function (CDF) for the distribution.

        :param x: Input array at which to evaluate the CDF.
        """
        pass

    def logcdf(self, x: np.ndarray) -> np.ndarray: # type: ignore[empty-body]
        r"""
        Compute the log cumulative density function (logCDF) for the distribution.

        :param x: Input array at which to evaluate the logCDF.
        """
        pass

    def stats(self) -> Dict[str, float]: # type: ignore[empty-body]
        r"""
        Computes and returns the statistical properties of the distribution, including,

        #. mean,
        #. median,
        #. variance, and
        #. standard deviation.

        :returns: A dictionary containing statistical properties such as mean, variance, etc.
        :rtype: Dict[str, float]

        Notes
        -----
        If any of the parameter is not computable for a distribution, this method returns None.
        """
        pass

    def _get_stats(self, key: str) -> Optional[float]:
        stats = self.stats()
        return stats.get(key) if stats else None

    @property
    def mean(self) -> Optional[float]:
        """The mean of the distribution."""
        return self._get_stats("mean")

    @property
    def median(self) -> Optional[float]:
        """The median of the distribution."""
        return self._get_stats("median")

    @property
    def mode(self) -> Optional[float]:
        """The mode of the distribution."""
        return self._get_stats("mode")

    @property
    def variance(self) -> Optional[float]:
        """The variance of the distribution."""
        return self._get_stats("variance")

    @property
    def stddev(self) -> Optional[float]:
        """The standard deviation of the distribution."""
        return self._get_stats("std")



"""Created on Aug 18 23:52:19 2024"""

__all__ = ["parameter_logic", "sanity_check", "_plot_fit"]

from typing import List, Tuple, Union, Optional, Callable

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from mpyez.backend.uPlotting import LinePlot # type: ignore
from mpyez.ezPlotting import plot_xy # type: ignore

from .. import ListOrNdArray

# SAFEGUARD:
xy_tuple = Tuple[np.ndarray, np.ndarray]
indexType = Union[int, List[int], None]


def sanity_check(
    x_values: ListOrNdArray,
    y_values: ListOrNdArray,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Convert input lists to NumPy arrays if necessary.

    Parameters
    ----------
    x_values : list of float or np.ndarray
        Input x-values that will be converted to a NumPy array if they are in list format.
    y_values : list of float or np.ndarray
        Input y-values that will be converted to a NumPy array if they are in list format.

    Returns
    -------
    x_values : np.ndarray
        The x-values as a NumPy array.
    y_values : np.ndarray
        The y-values as a NumPy array.
    """
    x_values = np.asarray(a=x_values, dtype=float)
    y_values = np.asarray(a=y_values, dtype=float)

    return x_values, y_values


def parameter_logic(
    par_array: np.ndarray,
    n_par: int,
    selected_models,
) -> np.ndarray:
    """
    Extract parameter values from a given function based on the number of parameters per fit and selected indices.

    Parameters
    ----------
    par_array : np.ndarray
        A 2D array where the first column contains the parameter values and the second contains its standard errors.
    n_par : int
        The number of parameters per fit (e.g., amplitude, mu, sigma, etc.).
    selected_models : int, list of int, or None
        Indices of model components to extract.
        - If None, selects all components.
        - If int or list of int, selects the specified components (1-based indexing).

    Returns
    -------
    np.ndarray
        A 2D array containing the selected parameter values for the specified mean and error values for the fit.
    """
    indices = np.array(selected_models) - 1 if selected_models is not None else slice(None)
    return par_array.reshape(-1, n_par)[indices]


def _plot_fit(
    x_values: np.ndarray,
    y_values: np.ndarray,
    parameters: np.ndarray,
    n_fits: int,
    class_name: str,
    _n_fitter: Callable,
    _n_plotter: Callable,
    show_individuals: bool = False,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    title: Optional[str] = None,
    data_label: Optional[str] = None,
    fit_label: Optional[str] = None,
    axis: Optional[Axes] = None,
):
    """
    Base function to plot the fitted models.

    Parameters
    ----------
    x_values : array-like
        The x-axis values.
    y_values : array-like
        The observed data values corresponding to `x_values`.
    parameters : tuple or list
        The optimized parameters from the fitting process.
    n_fits : int
        The number of fits performed.
    class_name : str
        The name of the fitting model class used.
    _n_fitter : callable
        A function that evaluates the fitted model given `x_values` and `parameters`.
    _n_plotter : callable
        A function that plots individual model components if `show_individuals` is True.
    show_individuals: bool, optional
        Whether to show individually fitted models or not.
    x_label: str, optional
        The label for the x-axis.
    y_label: str, optional
        The label for the y-axis.
    title: str, optional
        The title for the plot.
    data_label: str, optional
        The label for the data.
    axis: Axes, optional
        Axes to plot instead of the entire figure. Defaults to None.

    Returns
    -------
    plotter
        The plotter handle for the drawn plot.
    """
    if parameters is None:
        raise RuntimeError("Fit not performed yet. Call fit() first.")

    if data_label is None:
        dl, tt = "Data", "Total fit"
    elif len(data_label) == 1 or isinstance(data_label, str):
        dl, tt = data_label, "Total fit"
    elif 1 < len(data_label) <= 2:
        dl, tt = data_label, fit_label
    else:
        raise ValueError()

    plotter = plot_xy(
        x_data=x_values,
        y_data=y_values,
        data_label=dl,
        axis=axis,
        plot_dictionary=LinePlot(alpha=0.75),
    )

    plot_xy(
        x_data=x_values,
        y_data=_n_fitter(x_values, *parameters),
        x_label=x_label,
        y_label=y_label,
        plot_title=title,
        data_label=tt,
        plot_dictionary=LinePlot(color="k"),
        axis=plotter,
    )

    if show_individuals:
        _n_plotter(plotter=plotter)

    plotter2: Axes = plotter[0] if isinstance(plotter, list) else plotter

    plotter2.set_xlabel(x_label if x_label else "X")
    plotter2.set_ylabel(y_label if y_label else "Y")
    plotter2.set_title(title if title else f"{n_fits} {class_name} fit")
    plt.tight_layout()

    return plotter2



"""Created on Dec 04 03:57:18 2024"""

from typing import Dict

import numpy as np

from .backend import errorHandling as erH, BaseDistribution
from .utilities_d import half_normal_pdf_, half_normal_cdf_, half_normal_log_pdf_, half_normal_log_cdf_
from .. import md_scipy_like


class HalfNormalDistribution(BaseDistribution):
    r"""
    Class for halfnormal distribution.

    :param amplitude: The amplitude of the PDF. Defaults to 1.0. Ignored if **normalize** is ``True``.
    :type amplitude: float, optional

    :param scale: The standard deviation parameter, :math:`\sigma`. Defaults to 1.0.
    :type scale: float, optional

    :param loc: The location parameter, for shifting. Defaults to 0.0.
    :type loc: float, optional

    :param normalize: If ``True``, the distribution is normalized so that the total area under the PDF equals 1. Defaults to ``False``.
    :type normalize: bool, optional

    :raise NegativeAmplitudeError: If the provided value of amplitude is negative.
    :raise NegativeStandardDeviationError: If the provided value of standard deviation is negative.

    Examples
    --------
    Importing libraries:

    .. literalinclude:: ../../../examples/basic/halfnorm.py
       :language: python
       :linenos:
       :lineno-start: 3
       :lines: 3-7

    Generating a standard Half Normal(:math:`\sigma = 1`) distribution with ``pyMultiFit`` and ``scipy``:

    .. literalinclude:: ../../../examples/basic/halfnorm.py
       :language: python
       :linenos:
       :lineno-start: 9
       :lines: 9-12

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/halfnorm.py
       :language: python
       :linenos:
       :lineno-start: 14
       :lines: 14-29

    .. image:: ../../../images/half_normal_example1.png
       :alt: HN(1)
       :align: center

    Generating a translated Gaussian(:math:`\sigma=3`) distribution with :math:`\text{loc}=3`:

    .. literalinclude:: ../../../examples/basic/halfnorm.py
       :language: python
       :lineno-start: 32
       :lines: 32

    Plotting **PDF** and **CDF**:

    .. literalinclude:: ../../../examples/basic/halfnorm.py
       :language: python
       :lineno-start: 34
       :lines: 34-49

    .. image:: ../../../images/half_normal_example2.png
       :alt: HN(2, 3)
       :align: center
    """

    def __init__(
        self,
        amplitude: float = 1.0,
        scale: float = 1.0,
        loc: float = 0.0,
        normalize: bool = False,
    ):
        if not normalize and amplitude <= 0:
            raise erH.NegativeAmplitudeError()
        if scale < 0:
            raise erH.NegativeScaleError()
        self.amplitude = 1 if normalize else amplitude
        self.scale = scale
        self.loc = loc

        self.norm = normalize

    @classmethod
    @md_scipy_like('1.0.7')
    def scipy_like(cls, loc: float = 0.0, scale: float = 1.0) -> 'HalfNormalDistribution':
        """
        Instantiate HalfNormalDistribution with scipy parametrization.

        Parameters
        ----------
        loc: float, optional
            The location parameter. Defaults to 0.0.
        scale: float, optional
            The scale parameter. Defaults to 1.0.

        Returns
        -------
        HalfNormalDistribution
            An instance of normalized HalfNormalDistribution.
        """
        return cls(loc=loc, scale=scale, normalize=True)

    @classmethod
    def from_scipy_params(cls, loc: float = 0.0, scale: float = 1.0) -> 'HalfNormalDistribution':
        """
        Instantiate HalfNormalDistribution with scipy parametrization.

        Parameters
        ----------
        loc: float, optional
            The location parameter. Defaults to 0.0.
        scale: float, optional
            The scale parameter. Defaults to 1.0.

        Returns
        -------
        HalfNormalDistribution
            An instance of normalized HalfNormalDistribution.
        """
        return cls(loc=loc, scale=scale, normalize=True)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return half_normal_pdf_(
            x,
            amplitude=self.amplitude,
            sigma=self.scale,
            loc=self.loc,
            normalize=self.norm,
        )

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        return half_normal_log_pdf_(
            x,
            amplitude=self.amplitude,
            sigma=self.scale,
            loc=self.loc,
            normalize=self.norm,
        )

    def cdf(self, x: np.ndarray) -> np.ndarray:
        return half_normal_cdf_(
            x,
            amplitude=self.amplitude,
            sigma=self.scale,
            loc=self.loc,
            normalize=self.norm,
        )

    def logcdf(self, x: np.ndarray) -> np.ndarray:
        return half_normal_log_cdf_(
            x,
            amplitude=self.amplitude,
            sigma=self.scale,
            loc=self.loc,
            normalize=self.norm,
        )

    def stats(self) -> Dict[str, float]:
        s_, l_ = self.scale, self.loc

        mean_ = np.sqrt(2 / np.pi)
        mode_ = 0

        variance_ = 1 - (2 / np.pi)
        variance_ *= s_**2

        return {
            "mean": (s_ * mean_) + l_,
            "mode": mode_,
            "variance": variance_,
            "std": np.sqrt(variance_),
        }



"""Created on Jul 12 05:01:19 2025"""

import numpy as np
from numpy import ndarray

from . import BaseDistribution
from ..utilities_d import line, quadratic, cubic


class LineFunction(BaseDistribution):
    def __init__(
        self,
        slope: float = 1.0,
        intercept: float = 1.0,
        normalize: bool = False,
    ):
        self.slope = slope
        self.intercept = intercept

        self.norm = normalize

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Calculates the line function.

        Parameters
        ----------
        x : np.ndarray
            Input array of values.

        Returns
        -------
        np.ndarray
            Array of the same shape as :math:`x`, containing the evaluated values.
        """
        return line(x, slope=self.slope, intercept=self.intercept)


class QuadraticFunction(BaseDistribution):
    def __init__(
        self,
        a: float = 1.0,
        b: float = 1.0,
        c: float = 1.0,
        normalize: bool = False,
    ):
        self.a = a
        self.b = b
        self.c = c

        self.norm = normalize

    def pdf(self, x: ndarray) -> ndarray:
        """Calculates the quadratic function.

        Parameters
        ----------
        x : np.ndarray
            Input array of values.

        Returns
        -------
        np.ndarray
            Array of the same shape as :math:`x`, containing the evaluated values.
        """
        return quadratic(x, a=self.a, b=self.b, c=self.c)


class CubicFunction(BaseDistribution):
    def __init__(
        self,
        a: float = 1.0,
        b: float = 1.0,
        c: float = 1.0,
        d: float = 1.0,
        normalize: bool = False,
    ):
        self.a = a
        self.b = b
        self.c = c
        self.d = d

        self.norm = normalize

    def pdf(self, x: ndarray) -> ndarray:
        """Calculates the cubic function.

        Parameters
        ----------
        x : np.ndarray
            Input array of values.

        Returns
        -------
        np.ndarray
            Array of the same shape as :math:`x`, containing the evaluated values.
        """
        return cubic(x, a=self.a, b=self.b, c=self.c, d=self.d)
